from datetime import date
import subprocess as sp
import os
import sys


class latexdocument:
    """ A MARINTEK latex document (base class).
        Don't call directly. Use sub classes:
        sub class report
        sub class beamer
    """
    def __init__(self,name="Statoil.tex", title="Presentation",
                 subtitle="", author="Timothy E. Kendon", 
                 institute="FD GPU UDWP\\"
                            "\\Statoil, Trondheim",
                 date = date.today().isoformat(), subject=""):
        self.name      = name      # This is the filename
        self.title     = title     # This is the main report title
        self.subtitle  = subtitle  # This is the sub title
        self.institute = institute # This is the issuing institute
        self.author    = author    # This is the author
        self.date      = date      # This is the date
        self.subject   = subject   # This is the subject

    def open(self):
        """ Opens a new document and sets preamble.
            self.type is set by the subclass 
        """ 
        # Open document
        self.file = open(self.name, "w")
        # Write generic preamble 
        self.file.write(
            "\\documentclass[]{%s}\n"
            "\\usepackage[absolute,overlay]{textpos}\n"
            "\\usepackage{epsfig}\n"
            "\\usepackage{graphics}\n"
            "\\usepackage[english]{babel}\n"
            "\\usepackage{subfigure}\n"
            "\\usepackage{rotating}\n"
            "\\usepackage[]{hyperref}\n"
            #"\\usepackage{a4wide}\n"
            "\\usepackage{xcolor}\n"
            "\\usepackage{array}\n"
            "\\usepackage{multirow}\n"
            "\\usepackage{tabularx}\n"
            "\\usepackage{lscape}\n"
            "\\usepackage{pdfpages}\n"
            "\\usepackage[latin1]{inputenc}\n"
            "\\usepackage{paralist}\n"
            "\\newenvironment{myinparaenum}\n"
            "{\\begin{inparaenum}[\\hspace{2em}(a)]\\hspace{-2em}\\ignorespaces}\n"
            "{\\end{inparaenum}}\n"
            "\\usepackage{times}\n"
            "\\usepackage[T1]{fontenc}\n"
            "\\setcounter{secnumdepth}{3}\n"
            "\\setcounter{tocdepth}{3}\n" 
            "\\let\\oldtabular=\\tabular\n"
            "\\def\\tabular{\\footnotesize\\oldtabular}\n"
            % self.type)

    def close(self):
        self.file.write("\\end{document}")
        self.file.close()


class report(latexdocument):
    """ A MARINTEK report document """
    def __init__(self,name="MARINTEK", *args , **kwargs):
        name = name + "-Report-%s.tex" % date.today().isoformat()
        latexdocument.__init__(self, name, *args , **kwargs)

    def open(self):
        self.type = "report"
        latexdocument.open(self)       
        

class beamer(latexdocument):
    """ A MARINTEK beamer document """
    def __init__(self,name="MARINTEK", *args , **kwargs):
        name = name + "-Slides-%s.tex" % date.today().isoformat()
        latexdocument.__init__(self, name=name, *args , **kwargs)
        path = os.path.dirname(os.path.abspath( __file__ ))# + "\marintek02.pdf"
        path = "\\\\".join(path.split('\\'))
        sp.Popen('copy "%s\\\\frontpage_statoil_ppt.png" .' % path, shell=True)
        sp.Popen('copy "%s\\\\backpage_statoil_ppt.png" .'  % path, shell=True)
        sp.Popen('copy "%s\\\\statoil.png" .'               % path, shell=True)
        sp.Popen('copy "%s\\\\statoil_ppt.png" .'           % path, shell=True)

    def open(self):
        self.type = "beamer"
        latexdocument.open(self)
        self.file.write(
            "\\mode<presentation>{\n"
            "\\usetheme{Boadilla}\n"
            "\\setbeamercovered{transparent}}\n"
            "\\title[]{%s}\n"
            "\\subtitle{%s}\n"
            "\\author{%s}\n"
            "\\institute[]{%s}\n"
            "\\date[%s]{%s}\n"
            "\\subject{%s}\n"
            "\\AtBeginSubsection[]{\n"
            "\\begin{frame}<beamer>{Outline}\n"
            "\\tableofcontents[currentsection,currentsubsection,subsubsectionstyle=show/shaded]\n"
            "\\end{frame}}\n"
            "\setbeamertemplate{footline}[page number]{}\n"
            "\\setbeamertemplate{navigation symbols}{}\n"
            "\\usebackgroundtemplate{\n"
            "\\includegraphics[width=\\paperwidth,height=\\paperheight]{statoil_ppt.png}}\n"
            "\\newcommand\TitleText[1]{ \n"
            "\\begin{textblock*}{0.7\\paperwidth}(1em,0.8\\textheight)\n"
            "\\raggedright #1\\hspace{.5em}\n"
            "\\end{textblock*}}\n"
            "\\begin{document}\n"            
            "{\n"
            "\\usebackgroundtemplate{\\includegraphics[width=\\paperwidth]{frontpage_statoil_ppt.png}}\n"
            "\\begin{frame}\n"
            "\\TitleText{%s}"
            "\\end{frame}\n"
            "}\n"
            "\\begin{frame}{Outline}\n"
            "\\tableofcontents\n"
            "\\end{frame}\n" %
            (self.title,self.subtitle,self.author,
             self.institute,self.date,self.date,self.subject, self.title))


    def addSection(self,name):
        self.file.write("\\section{%s}\n" % name)
        
    def addSubsection(self,name):
        self.file.write("\\subsection{%s}\n" % name)

    def addSubsubsection(self,name):
        self.file.write("\\subsubsection{%s}\n" % name)
            
    def addFrame(self,title,subtitle,imagelist=None,bullets=None):
        self.file.write(
            "\\begin{frame}{%s}{%s}\n" % (title, subtitle))
        if imagelist is not None:
            for im in imagelist: self.file.write(
                "\\includegraphics[width=1.0\\textwidth]{%s}" % im)
        self.file.write("\\end{frame}\n")
            

    def addBackPage(self):
        self.file.write(
            "{\n"
            "\\usebackgroundtemplate{\\includegraphics[width=\\paperwidth]{backpage_statoil_ppt.png}}\n"
            "\\begin{frame}\n"
            "\\end{frame}\n"
            "}\n")



def main():
    """ Test Program for routines in latex.py """
    wamdoc = beamer(name = "DiffractionAnalysis",
                    title= "Diffraction Analysis for the Njord A TLP")
    wamdoc.open()
    wamdoc.addSection("1st Order Diffraction Analysis")
    Modes = ["Surge","Sway","Heave","Roll","Pitch","Yaw"]

    # Add Wave Excitation Coefficients
    wamdoc.addSubsection("Wave Excitation Force Coefficients")
    for mode in Modes:
        wamdoc.addFrame("Wave Excitation Force Coefficients", 
                        "Platform %s Direction" % mode,
                        imagelist=["figures/Fexc_D_%s.png" % mode])

    # Add R.A.Os
    wamdoc.addSubsection("Response Amplitude Operators")
    for mode in Modes:
        wamdoc.addFrame("Wave Excitation Coefficients", 
                        "Platform %s Direction" % mode,
                        imagelist=["figures/Response_%s.png" % mode])


    # Add Mean Wave Drift Coefficients
    wamdoc.addSubsection("Mean Wave Drift Force Coefficients")
    for mode in Modes:
        wamdoc.addFrame("Mean Wave Drift Force Coefficients", 
                        "Platform %s Direction" % mode,
                        imagelist=["figures/Drift_P_%s.png" % mode])


    # Add Added Mass or Damping Coefficients
    wamdoc.addSubsection("Added Mass and Damping Coefficients")
    Modes = ["11", "15", "22", "24", "33", "42", "44","51","55","66"]
    for mode in Modes:
        wamdoc.addFrame("Added Mass Coefficients", 
                        "A%s" % mode, 
                        imagelist=["figures/A%s.png" % mode])
    for mode in Modes:
        wamdoc.addFrame("Damping Coefficients", 
                        "B%s" % mode, 
                        imagelist=["figures/B%s.png" % mode])

    wamdoc.close()





import sys
if __name__ == '__main__':
    sys.exit(main())



import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import importlib.resources as pkg_resources


def mehaute(wave_params=None):
    """
    Wave theory categorization plot
    wave_params(H, h, T)
    H - wave height
    h - water depth
    T - wave period
    """

    # Setup plot
    fig, ax = plt.subplots(1,1,figsize=(8,6))
    ax.set_xscale("log")
    ax.set_yscale("log")

    # Read the data
    fname = os.path.join(os.path.dirname(__file__), "mehaute.txt")
    x, y = np.loadtxt(fname, delimiter=",").T # data points

    # Shallow water and deep water boundary lines
    shallow_water, deep_water = x[2:4]

    # Linear wave theory
    linear_x = np.r_[x[0],x[6:28]]
    linear_y = np.r_[y[0],y[6:28]]
    # Stokes 2nd order
    second_x = np.r_[x[28:42]]
    second_y = np.r_[y[28:42]]
    # Stokes 3rd order
    third_x = np.r_[x[4],x[42:53]]
    third_y = np.r_[y[4],y[42:53]]
    # Stokes 4th/5th order
    fourth_x = np.r_[x[5],x[53:60]]
    fourth_y = np.r_[y[5],y[53:60]]
    # breaking criteria
    break_x = np.r_[x[0],x[1]]
    break_y = np.r_[y[0],y[1]]    
    # plot 
    ax.set_xlim(0.0001, 1.0)
    ax.set_ylim(0.00001, 0.1)
    ax.plot(linear_x, linear_y)
    ax.plot(second_x, second_y)
    ax.plot(third_x, third_y)
    ax.plot(fourth_x, fourth_y)
    ax.plot(break_x, break_y)
    ax.axvline(shallow_water, linestyle='--')
    ax.axvline(deep_water, linestyle='--')
    ax.text(0.00015, 0.11, "Shallow water")
    ax.text(0.003, 0.11, "Intermediate depth")
    ax.text(0.13, 0.11, "Deep water")
    ax.text(0.13, 0.0001, "Linear")
    ax.text(0.13, 0.003, "Stokes 2nd")
    ax.text(0.13, 0.013, "Stokes 3rd")
    ax.text(0.0002, 0.00005, "Cnoidal")
    ax.text(0.00048, 0.000335, "Solitary", rotation = 35)
    ax.set_xlabel("$\dfrac{h}{gT^2}$")
    ax.set_ylabel("$\dfrac{H}{gT^2}$", rotation = 0)

    if wave_params is not None:
        H, T, h = wave_params
        g = 9.80665
        x = h/g/T**2
        y = H/g/T**2
        print(x, y)
        ax.scatter(x, y, label="Conditions")

    plt.show()



if __name__ == "__main__":
    H = 4 #np.r_[4, 16] 
    T = 10 #np.r_[10, 20]
    h = 1000
    mehaute(wave_params=(H, T, h))


# -*- coding: utf-8 -*-
"""
 Common Files
"""


class Record(object):
    """ A record class that allows you to automatically
    generate a structure based on strings.
    """

    def __init__(self, name):
        self.name = name

    def setattr(self, name, value):
        parts = name.split(".")
        obj = self
        for attr in parts[:-1]:
            if hasattr(obj, attr):
                obj = getattr(obj, attr)
                if not isinstance(obj, Record):
                    raise AttributeError
            else:
                setattr(obj, attr, Record(attr))
                obj = getattr(obj, attr)
        setattr(obj, parts[-1], value)

    def addattr(self, **kwargs):
        for k, v in kwargs.items():
            self.setattr(k, v)



def loadtxt(filename, transpose=False, skip_header=0, usecols=None):
    """
    Load text with string and number datatypes. 
    Returns column data as row data for unpacking
    """
    import numpy
    raw_data = numpy.genfromtxt(filename, dtype=None, encoding=None, skip_header=skip_header, 
                                usecols=usecols)

    if transpose:
        curated_data = [numpy.array(arr) for arr in zip(*raw_data)]
        for ivar, row in enumerate(curated_data):
            if isinstance(row[0], numpy.bytes_):
                curated_data[ivar] = numpy.r_[[item.decode("utf-8") for item in row]]

    else:
        curated_data = []
        for row in raw_data:
            curated_data.append([item.decode("utf-8") if isinstance(item, numpy.bytes_) else item for item in row])
    return curated_data


#
# Script: Reads RAOs from either SIMO or WAMIT
#

import numpy as np
from ..miscellaneous.io import loadtxt
from ..administration.tools import find_occurrences, timer
from ..miscellaneous.math import interpc
from .common import Record
import sys
import enum 
from abc import ABC, abstractclassmethod
import scipy.optimize as opt
import time 
import os
import typing



class Wamit(object):
    """ Class for working with WAMIT files
    TODO: Implement an effective read (note we can have static functions)
    //TODO: Implement a wave to wave TF transformation to field location
    TODO: Implement a wave to body motion TF transformation to field location
    TODO: Implement a morison element type for general shapes e.g. circles, lines
    //TODO: Implement the EOM   
    TODO: Set external quadratic damping matrices
    //# TODO: Set external stiffness matrix

    Assumes consistent periods in numerical files (no mixing)
    Implemented for one body only
    Phases of exciting forces, motions, pressure and fluid velocity are defined
    relative to the incident-wave elevation at X=Y=0 which is equal to Re(A*e(iwt))=Acos(wt)
    """

    class Motion(enum.IntEnum):
        SURGE = 0
        SWAY = 1
        HEAVE = 2
        ROLL = 3
        PITCH = 4
        YAW = 5

    class Properties(enum.Enum):
        PERIODS = enum.auto()
        DIRECTIONS = enum.auto()
        MASS = enum.auto()
        ADDED_MASS = enum.auto()
        ADDED_MASS_INF = enum.auto()
        ADDED_MASS_ZERO = enum.auto()
        DAMPING = enum.auto()
        EXCITE_HASK = enum.auto()
        EXCITE_DIFF = enum.auto()
        RAOS = enum.auto()
        BODY_PANELS = enum.auto()
        BODY_PRESSURE_TRANSFER_FUNCTIONS = enum.auto()
        STIFFNESS_HYDROGRAV = enum.auto() 
        STIFFNESS_EXTERNAL = enum.auto() 
        DAMPING_EXTERNAL = enum.auto()
        QUADRATIC_DAMPING_EXTERNAL = enum.auto()
        LINEARIZED_QUADRATIC_DAMPING_EXTERNAL = enum.auto() # purely internal
        ADDED_MASS_ZERO_EXTERNAL = enum.auto()
        RETARDATION_FUNCTIONS = enum.auto()
        SLENDER_ELEMENTS = enum.auto()
        NATURAL_PERIODS = enum.auto()
        CRITICAL_DAMPING = enum.auto()
        DISPLACED_VOLUMES = enum.auto()
        MASS_CENTER = enum.auto()
        BUOYANCY_CENTER = enum.auto()
        BODY_POINTS = enum.auto()


    class SlenderElement(object):
        def __init__(self, x1, x2, xref, Cl, Cq, Ca, N):
            """
                Cl - Linear drag coefficients in x, y, z (R^3)
                Cq - Quadratic drag coefficients in x, y, z (R^3)
                Ca - Added mass coefficient in x, y, z (R^3)
            """
            self._x1 = x1
            self._x2 = x2
            self._xref = xref
            self._Cl = Cl
            self._Cq = Cq
            self._Ca = Ca
            # Segment lengths
            self._dl = np.linalg.norm(x2 - x1)/N
            # Segment centroid coordinates 
            self._xp = np.array([x1 + (x2-x1)*(i+0.5)/N for i in range(N)])
            # Compute local unit basis
            xm = x2 - x1
            zm = np.cross(xm, xref-x1)
            ym = np.cross(zm, xm)
            # Compute local basis
            self._ex = xm/np.linalg.norm(xm)
            self._ey = ym/np.linalg.norm(ym)
            self._ez = zm/np.linalg.norm(zm)
            # Compute projection matrix for the perpendicular plane
            xn, yn, zn = self._ex
            self._Pk = np.array([
                [yn**2+zn**2, -xn*yn, -xn*zn],
                [-yn*xn, xn**2+yn**2, -yn*zn],
                [-zn*xn, -zn*yn, xn**2+yn**2]])



        def _compute_force(self, u, A=1.0):
            """ u is a velocity field at points self._xp of size (Np,3), where
                u(xp,:) = velocity of incident at xp wave for RHS term (viscous excitation)
                u(xp,:) = velocity of motion at xp for LHS term (viscous damping)
                xp(Np,3) = body coordinates of points
            """
            e = self.local_basis 
            B = self._Cq*8/3/np.pi*A*(2*np.pi/10)

            xp = self.points
            dl = self._dl
            u = np.einsum("ijkl->klij", u)  # Np,3,nBeta,nPeriod => nBeta,nPeriod,Np,3
            dF = (u@(e.T)*B*dl)@e  # shape nBeta,nPeriod,Np,3
            dM = np.cross(xp, dF, axisa=1, axisb=3) # nBeta,nPeriod,Np,3
            F = np.sum(dF, axis=2) # sum over the 3rd axis (==2nd zero indexed)
            M = np.sum(dM, axis=2) # sum over the 3rd axis (==2nd zero indexed)
            F = np.einsum("ijk->kij", F)
            M = np.einsum("ijk->kij", M)
            F = np.vstack([F, M])
            return F



        @property
        def local_basis(self): return np.vstack([self._ex,self._ey,self._ez])

        @property
        def points(self): return self._xp


    @timer
    def __init__(self, fname, ulen=1.0, density=1.025, g=9.80665, 
                frequency_ascending=False, verbose=False):
        self.fname = fname 
        self._data = {}
        for prop in Wamit.Properties: self._data[prop] = None 
        if os.path.isfile(f"{fname}.1"):
            # Read and unpack added mass (assumes .1 is always there)
            wam = Wamit.read_added_mass_and_damping(f"{fname}.1", ulen, density, frequency_ascending)
            self._data[Wamit.Properties.PERIODS] = wam.periods
            self._data[Wamit.Properties.ADDED_MASS] = wam.added_mass
            self._data[Wamit.Properties.DAMPING] = wam.damping
            self._data[Wamit.Properties.ADDED_MASS_INF] = wam.ainf
            self._data[Wamit.Properties.ADDED_MASS_ZERO] = wam.a0
        if os.path.isfile(f"{fname}.2"):
            # Read and unpack excitation forces from Haskind relations
            wam = Wamit.read_excitation_forces(f"{fname}.2", ulen, density, g, frequency_ascending)
            self._data[Wamit.Properties.EXCITE_HASK] = wam.value
            self._data[Wamit.Properties.DIRECTIONS] = wam.directions
        if os.path.isfile(f"{fname}.3"):
            # Read and unpack excitation forces diffraction potential
            wam = Wamit.read_excitation_forces(f"{fname}.3", ulen, density, g, frequency_ascending)
            self._data[Wamit.Properties.EXCITE_DIFF] = wam.value
            self._data[Wamit.Properties.DIRECTIONS] = wam.directions
        if os.path.isfile(f"{fname}.4"):
            # Read and unpack motion transfer functions
            wam = Wamit.read_raos(f"{fname}.4", ulen, 1, frequency_ascending)
            self._data[Wamit.Properties.RAOS] = wam.value 
        if os.path.isfile(f"{fname}.5p"): # TODO currently assumes ulen=1
            wam = Wamit.read_body_pressure_tfs(f"{fname}.5p")
            self._data[Wamit.Properties.BODY_PRESSURE_TRANSFER_FUNCTIONS] = \
                {"diffraction": wam.h_dif, "radiation": wam.h_rad}
        if os.path.isfile(f"{fname}.pnl"): # TODO currently assumes ulen=1
            wam = Wamit.read_wamit_normals(f"{fname}.pnl") 
            self._data[Wamit.Properties.BODY_PANELS] = wam
        if os.path.isfile(f"{fname}.mmx"):
            # Read and unpack mass data
            wam = Wamit.read_mass_matrix(f"{fname}.mmx")
            self._data[Wamit.Properties.MASS] = wam.mass
            self._data[Wamit.Properties.DAMPING_EXTERNAL] = wam.damping
            self._data[Wamit.Properties.STIFFNESS_EXTERNAL] = wam.stiffness
            self._data[Wamit.Properties.DISPLACED_VOLUMES] = wam.vols
            self._data[Wamit.Properties.MASS_CENTER] = wam.cog
            self._data[Wamit.Properties.BUOYANCY_CENTER] = wam.cob
        if os.path.isfile(f"{fname}.out"):
            # Read stiffness data from out file
            wam = Wamit.read_hydrograv_stiffness(f"{fname}.out", ulen, density, g)
            self._data[Wamit.Properties.STIFFNESS_HYDROGRAV] = wam.stiffness
        # Evaluate retardation functions
        #self._update_retardation_functions()
        # Evaluate natural periods in heave, roll and pitch
        self._compute_natural_periods(verbose)
        # Initial storage for quadratic external damping
        self._data[Wamit.Properties.QUADRATIC_DAMPING_EXTERNAL] = np.zeros((6,6))
        # Initial storage for added mass adjustment
        self._data[Wamit.Properties.ADDED_MASS_ZERO_EXTERNAL] = np.zeros((6,6))
        self._data[Wamit.Properties.LINEARIZED_QUADRATIC_DAMPING_EXTERNAL] = np.zeros((len(self.directions),6,6))


    @property
    def registered_body_points(self):
        assert self._data[Wamit.Properties.BODY_POINTS] is not None, "No body points registered. Use register_body_points"
        return self._data[Wamit.Properties.BODY_POINTS]

    def register_body_points(self, xyz, clean=False):
        """ 
        Register the a numpy array of xyz body point data for the computation of statistics
        Point data given in rows. If just a single point, the xyz is a row vector. 
        The clean option replaces existing points with the new dataset, otherwise
        the points are appended to the existing array of points
        """
        if clean or self._data[Wamit.Properties.BODY_POINTS] is None:
            if xyz.ndim == 1: xyz = np.expand_dims(xyz, axis=0)
            self._data[Wamit.Properties.BODY_POINTS] = xyz
        else:
            self._data[Wamit.Properties.BODY_POINTS] = \
                np.vstack([self._data[Wamit.Properties.BODY_POINTS], xyz])


    def get_raos(self, mode=None, direction=None, return_at_body_points=False,
                return_velocity=False, return_acceleration=False, relative=False):
        """
            Returns the stored wave to motion transfer functions, which need to be precalculated
            using either the solve_EOM function or the solve_stochastic function. 
            The user can request:
                - a specific mode of motion ... all modes of motion returned as default
                - a specific wave direction ... all directions return as default
                - acceleration or velocity to be return instead of the motion raos
            If return_at_body_points=True, the raos are returned at the body point locations 
            The body point locations are registered with register_body_points function.
            if returning at body point locations, only translation modes 0,1,2 available
        """

        # Make assertions
        if direction is not None:
            assert direction in self.directions, "direction not computed, direction interpolation not implemented"    
        if return_at_body_points:
            assert self._data[Wamit.Properties.BODY_POINTS] is not None, "need to register_body_points first"
            if mode is not None:
                assert mode in [Wamit.Motion.SURGE, Wamit.Motion.SWAY, Wamit.Motion.HEAVE], "Need to request Wamit.Motion.SURGE/SWAY/HEAVE"

        nmodes = 3 if return_at_body_points else 6    
        ix = range(nmodes) if mode is None else [mode,]
        jx = range(len(self.directions)) if direction is None else self.directions==direction
        ij = np.ix_(ix, jx)

        if return_at_body_points:
            assert self._data[Wamit.Properties.BODY_POINTS] is not None, "need to register_body_points first"
            X = self.transform_rao(self._data[Wamit.Properties.BODY_POINTS], 
                                    return_velocity=return_velocity, 
                                    return_acceleration=return_acceleration)
            if relative:
                # evaluate the incident wave motion RAOs
                wave = Wamit.compute_incident_wave(self.periods, self.directions, 
                                        self._data[Wamit.Properties.BODY_POINTS],
                                        return_velocity=return_velocity, return_acceleration=return_acceleration)
                X -= wave

            if len(self.registered_body_points) == 1:
                return np.expand_dims(np.squeeze(X[:, ix, jx]), axis=0)
            else: return np.squeeze(X[:, ix, jx])
        else:
            X = self._data[Wamit.Properties.RAOS]
            #print("here", ij, X[ij].shape)
            jw = 1j*(2*np.pi/self.periods)
            factor = jw if return_velocity else jw**2 if return_acceleration else 1            
            if relative:
                # evaluate the incident wave motion RAOs
                wave = Wamit.compute_incident_wave(self.periods, self.directions, 
                                        np.expand_dims(np.r_[0,0,0], axis=0), # origin
                                        return_velocity=return_velocity, return_acceleration=return_acceleration)
                wave = np.squeeze(wave) # only xyz components, so lets pad for other modes to make general
                wave = np.vstack([wave, np.zeros(wave.shape)])
                #print("wave shape: ", wave.shape)
            return np.squeeze(X[ij]*factor - wave[ij]) if relative else np.squeeze(X[ij]*factor)


    def get_pressures(self, direction, ulen=1):
        """
        """
        # Make assertions
        assert direction in self.directions, "direction not computed, direction interpolation not implemented"    
        pressure = self._data[Wamit.Properties.BODY_PRESSURE_TRANSFER_FUNCTIONS]
        p_dif = pressure["diffraction"]
        p_rad = pressure["radiation"]
        ix = np.nonzero(self.directions == direction)
        p_dif = p_dif[ix,:,:]
        return np.squeeze(p_dif), p_rad
        

    def compute_std_deviations(self, Sxx, fz, mode=None, direction=None, 
        return_at_body_points=False, return_velocity=False, return_acceleration=False, 
        relative=False):
        """
            Returns the standard deviation of response in a given sea spectrum Sxx(fz),
            where fz is in Hzs. This function uses pre-existing RAO data; to use this 
            function either solve_EOM or solve_stochastic must have been run. 
            The user can request:
                - a specific mode of motion ... all modes of motion returned as default
                - a specific wave direction ... all directions return as default
                - acceleration or velocity to be return instead of the motion raos
            If return_at_body_points=True, the values are returned at the body point locations 
            The body point locations are registered with register_body_points function.
            if returning at body point locations, only translation modes 0,1,2 available
        """
        
        X = self.get_raos(mode=mode, direction=direction, return_at_body_points=return_at_body_points,
                return_velocity=return_velocity, return_acceleration=return_acceleration, relative=relative)

        if X.ndim == 1:
            H = interpc(fz, 1/self.periods, X)
            return np.trapz(Sxx*np.abs(H)**2, fz)**0.5
        else:
            original_shape = X.shape
            size = int(np.product(original_shape))
            nfreqs = original_shape[-1] # also len(self.periods)
            X = X.reshape((size//nfreqs, nfreqs))
            stds = np.zeros(size//nfreqs)
            for i, Xi in enumerate(X):
                H = interpc(fz, 1/self.periods, Xi)
                stds[i] = np.trapz(Sxx*np.abs(H)**2, fz)**0.5
            return stds.reshape(original_shape[:-1])


    def compute_upcrossing_periods(self, Sxx, fz, mode=None, direction=None, 
        return_at_body_points=False, return_velocity=False, return_acceleration=False, 
        relative=False):
        """
            Returns the standard deviation of response in a given sea spectrum Sxx(fz),
            where fz is in Hzs. This function uses pre-existing RAO data; to use this 
            function either solve_EOM or solve_stochastic must have been run. 
            The user can request:
                - a specific mode of motion ... all modes of motion returned as default
                - a specific wave direction ... all directions return as default
                - acceleration or velocity to be return instead of the motion raos
            If return_at_body_points=True, the values are returned at the body point locations 
            The body point locations are registered with register_body_points function.
            if returning at body point locations, only translation modes 0,1,2 available
        """
        
        X = self.get_raos(mode=mode, direction=direction, return_at_body_points=return_at_body_points,
                return_velocity=return_velocity, return_acceleration=return_acceleration, relative=relative)

        if X.ndim == 1:
            H = interpc(fz, 1/self.periods, X)
            m0 = np.trapz(Sxx*np.abs(H)**2, fz)
            m2 = np.trapz(Sxx*np.abs(H)**2*fz**2, fz)
            return np.sqrt(self._suppressed_division(m0, m2))
        else:
            original_shape = X.shape
            size = int(np.product(original_shape))
            nfreqs = original_shape[-1] # also len(self.periods)
            X = X.reshape((size//nfreqs, nfreqs))
            Tz = np.zeros(size//nfreqs)
            for i, Xi in enumerate(X):
                H = interpc(fz, 1/self.periods, Xi)
                m0 = np.trapz(Sxx*np.abs(H)**2, fz)
                m2 = np.trapz(Sxx*np.abs(H)**2*fz**2, fz)
                Tz[i] = np.sqrt(self._suppressed_division(m0, m2))
            return Tz.reshape(original_shape[:-1])


    def compute_statistics(self, Sxx, fz, mode=None, direction=None, 
        return_at_body_points=False, return_velocity=False, return_acceleration=False, 
        relative=False, duration_of_seastate=3, user_specified_fractile=0.9):
        """
            Returns the standard deviation of response in a given sea spectrum Sxx(fz),
            where fz is in Hzs. This function uses pre-existing RAO data; to use this 
            function either solve_EOM or solve_stochastic must have been run. 
            The user can request:
                - a specific mode of motion ... all modes of motion returned as default
                - a specific wave direction ... all directions return as default
                - acceleration or velocity to be return instead of the motion raos
            If return_at_body_points=True, the values are returned at the body point locations 
            The body point locations are registered with register_body_points function.
            if returning at body point locations, only translation modes 0,1,2 available
        """
        
        X = self.get_raos(mode=mode, direction=direction, return_at_body_points=return_at_body_points,
                return_velocity=return_velocity, return_acceleration=return_acceleration, relative=relative)

        D  = duration_of_seastate*3600 # seconds

        #print("RAO stats shape: ", X.shape)

        if X.ndim == 1:
            H = interpc(fz, 1/self.periods, X)
            m0 = np.trapz(Sxx*np.abs(H)**2, fz)
            m2 = np.trapz(Sxx*np.abs(H)**2*fz**2, fz)
            Tz = np.sqrt(self._suppressed_division(m0, m2))
            N  = D/Tz
            #alpha = 1 - user_specified_fractile
            #user_extreme = np.sqrt(-2*m0*np.log(1.-(1.-alpha)**(1./N)))
            user_extreme = np.sqrt(-2*m0*np.log(1.-user_specified_fractile**(1./N)))
            mpm_extreme =  np.sqrt(2*m0*np.log(N))
            return np.sqrt(m0), Tz, mpm_extreme, user_extreme
        else:
            original_shape = X.shape
            size = int(np.product(original_shape))
            nfreqs = original_shape[-1] # also len(self.periods)
            X = X.reshape((size//nfreqs, nfreqs))
            tzs = np.zeros(size//nfreqs)
            stds = np.zeros(size//nfreqs)
            mpm_extremes = np.zeros(size//nfreqs)
            user_extremes = np.zeros(size//nfreqs)
            for i, Xi in enumerate(X):
                H = interpc(fz, 1/self.periods, Xi)
                m0 = np.trapz(Sxx*np.abs(H)**2, fz)
                m2 = np.trapz(Sxx*np.abs(H)**2*fz**2, fz)
                tzs[i] = np.sqrt(self._suppressed_division(m0, m2))
                N  = D/tzs[i]
                user_extremes[i] = np.sqrt(-2*m0*np.log(1.-user_specified_fractile**(1./N)))
                mpm_extremes[i] =  np.sqrt(2*m0*np.log(N))
                stds[i] = np.sqrt(m0)
            return np.array([stds.reshape(original_shape[:-1]), 
                    tzs.reshape(original_shape[:-1]),
                    mpm_extremes.reshape(original_shape[:-1]),
                    user_extremes.reshape(original_shape[:-1])])


    def _suppressed_division(self, nominator, demoninator):
        import warnings
        warnings.filterwarnings("ignore")
        return np.nan_to_num(nominator/demoninator)


    def _compute_natural_periods(self, verbose=False):
        if self._data[Wamit.Properties.STIFFNESS_HYDROGRAV] is None:
            raise Exception(f"Hydro stiffness not read. No {self.fname}.out file present")

        natural_periods = np.zeros(6) # storage for natural periods
        critical_damping = np.zeros(6) # storage for critical damping
        M = self._data[Wamit.Properties.MASS]   
        AM = self._data[Wamit.Properties.ADDED_MASS]   
        C = self._data[Wamit.Properties.STIFFNESS_HYDROGRAV]   
        Xg,Yg,Zg = self._data[Wamit.Properties.MASS_CENTER]
        w = 2*np.pi/self.periods
        
        
        # Formula for natural period relates to a COG centered system

        # Subtract of Steiner term to get inertia to COG
        for i in [3,4]: M[i,i] -= M[0,0]*Zg**2

        # TODO: How to adjust for added mass (rotation FS => COG)

        # Natural frequency in heave, roll and pitch
        for i in [2,3,4]:
            print(i, C[i,i], M[i,i], 2*np.pi*np.sqrt(M[i,i]/(C[i,i])))
            res = opt.minimize(
                fun=lambda wn: abs(wn - np.sqrt(C[i,i]/(M[i,i] + np.interp(wn,w,AM[i,i])))),
                x0=np.sqrt(C[i,i]/M[i,i]))

        #for i in [2,3,4]:
        #    print(i, C[i,i], M[i,i], 2*np.pi*np.sqrt(M[i,i]/(C[i,i])))
        #    res = opt.minimize(
        #        fun=lambda wn: abs(wn - np.sqrt(C[i,i]/(M[i,i]))),
        #        x0=np.sqrt(C[i,i]/M[i,i]))
            natural_periods[i] = 2*np.pi/res.x
            critical_damping[i] = 2*res.x*(M[i,i])
            AMii = np.interp(res.x,w,AM[i,i]) 
            print(AMii, AMii/M[i,i])

        self._data[Wamit.Properties.NATURAL_PERIODS] = natural_periods
        self._data[Wamit.Properties.CRITICAL_DAMPING] = critical_damping 
        
        if verbose:
            for i, dof in zip([2,3,4],"heave roll pitch".split()):
                wn = 2*np.pi/natural_periods[i]
                Tn = 2*np.pi/np.sqrt(C[i,i]/(M[i,i] + np.interp(wn,w,AM[i,i])))
                print(f"Computed natural periods in {dof}: {natural_periods[i]:.1f} s " +
                      f"(Check: {Tn:.1f} s)")


    def add_slender_element(self, x1, x2, xref, Cl, Cq, Ca, N):
        se = Wamit.SlenderElement(x1, x2, xref, Cl, Cq, Ca, N)
        if self._data[Wamit.Properties.SLENDER_ELEMENTS] is None:
            self._data[Wamit.Properties.SLENDER_ELEMENTS] = [se]
        else:
            self._data[Wamit.Properties.SLENDER_ELEMENTS] += [se]

    def _evaluate_slender_element_wave_excitation(self, iper, ibeta, amp=1.0, vstd=None, 
                                                    g=9.80665, verbose=False):
        """
            vstd relates to the 3d space (there is no vstd for the roll in this setup)
        """
        Fv = np.zeros(3, np.complex)
        Mv = np.zeros(3, np.complex)

        period = self.periods[iper]
        beta = self.directions[ibeta]
        omega = 2*np.pi/period
        k = omega**2/g

        
        for ise, se in enumerate(self._data[Wamit.Properties.SLENDER_ELEMENTS]):

            if vstd is not None:
                B = np.array([se._Cq*np.sqrt(8/np.pi)*vstd]).T # column vector (3,1)
            else:
                B = np.array([se._Cq*8/3/np.pi*amp*(2*np.pi/period)]).T # column vector (3,1)
                #B = se._Cq*8/3/np.pi*amp*(2*np.pi/period) # surely this needs commenting out 
            
            #print(f"Contribution from Slender Element {ise+1}")
            e = se.local_basis 
            points = se.points
            dl = se._dl           

            for ipoint, point in enumerate(points):       
                x,y,z = point
                r = np.array( # cross product matrix
                    [
                        [ 0, -z,  y],
                        [ z,  0, -x],
                        [-y,  x,  0]
                    ]
                )

                eta = np.exp(-1j*(k*x*np.cos(beta) + k*y*np.sin(beta))) # phase shift to x
                u  = np.r_[     # Wave velocity at the point
                    omega*np.cos(beta)*np.exp(k*z)*eta,
                    omega*np.sin(beta)*np.exp(k*z)*eta,
                    1j*omega*np.exp(k*z)*eta]
                print(f"period: {period}, u: {u}")
                u = -u  # for long wave expect it to be approx Fv and u to be ~ 0 - |F|*1j 
                dF = e.T@(B*dl*e)@u
                dM = np.cross(point, dF) #r@Fv
                Fv += dF
                Mv += dM


                print(f"Point: {point}")
                print(f"Fv contribution: {dF}")
                print(f"Mv contribution: {dM}")
            
        F = np.r_[Fv,Mv]
        print(f"Viscous Wave Force:\n{F}")
        print(f"Wave heading: {beta}")
        print(f"Wave period: {period}")
        sys.exit()
        return F


    def _evaluate_slender_element_damping(self, iper, ibeta, amp=1.0, vstd=None):
        """ Direction independent at the moment, and if vstd used, it is also
        period independent. So if vstd is used then the damping matrix is constant
        for all directions and periods which is desirable as a damping matrix is
        direction independent (direction is built into the rows and columns)
        """
        BV = np.zeros((6,6))
        for ise, se in enumerate(self._data[Wamit.Properties.SLENDER_ELEMENTS]):
            #print(f"Contibution from Slender Element {ise+1}")
            # u_I and u-B have dimensions (Np, 3, Nbeta, Nfreq)
            #u_I = Wamit.compute_incident_wave(self.periods, self.directions, se.points) 
            #u_B = self.transform_rao(se.points, return_velocity=True)
            e = se.local_basis 
            points = se.points
            dl = se._dl           
            # simplify matrix operations during initial implementation
            # by removing periods and directions from the matrix system
            period = self.periods[iper]
            beta = self.directions[ibeta]
            if vstd is not None:
                B = np.array([se._Cq*np.sqrt(8/np.pi)*vstd]).T # column vector (3,1)
            else:
                B = np.array([se._Cq*8/3/np.pi*amp*(2*np.pi/period)]).T # column vector (3,1)
            I = np.eye(3)
            O = np.zeros((3,3))
            
            for ipoint, point in enumerate(points):       
                x,y,z = point

                # Need to be careful with the application or r
                # For Force to Moment we have r ^ F
                # However for rotation to linear velocity we have theta ^ r
                r = np.array( # cross product matrix
                    [
                        [ 0, -z,  y],
                        [ z,  0, -x],
                        [-y,  x,  0]
                    ]
                )


                #print(f"local basis e:\n {e}")
                #print(f"Damping coeffs B:\n {B}")
                #print(f"Length of segment dl: {dl}")             
                #print(f"local basis B*dl*e:\n {B*dl*e}")
                #print(f"B.shape: {B.shape}")
                
                E = (B*dl*e)
                U1 = I # translation
                U2 = -r # rotation contribution (need to make negative as u = theta ^ r not r ^ theta)


                # Why bother with this transformation??? 
                #A11 = (U1.T@E.T).T  # top left block  => dF local contribution due to translations
                #A12 = (U2.T@E.T).T  # top right block => dF local contribution due to rotations
                A11 = E@U1 # top left block  => dF local contribution due to translations
                A12 = E@U2 # top right block => dF local contribution due to rotations


                #print(f"A11 (local):\n{A11}")
                #print(f"A12 (local):\n{A12}")

                A11 = e.T@A11  # top left block  => dF global contribution due to translations
                A12 = e.T@A12  # top right block => dF global contribution due to rotations

                #print(f"A11 (global):\n{A11}")
                #print(f"A12 (global):\n{A12}")
                
                # Not quite mastered this (needed -ve to make positive)
                A21 = r@A11  # top left block  => dM global contribution due to translations
                A22 = r@A12  # top right block => dM global contribution due to rotations

                #print(f"A21 (global):\n{A21}")
                #print(f"A22 (global):\n{A22}")

                A = np.block([[ A11, A12],[A21,  A22]])
                #print(f"block damping matrix contribution, A:\n {A}")
                BV += A

        #print(f"Summed, damping matrix:\n {BV}")
        #sys.exit()
        return BV
           

    def transform_rao(self, body_pts, return_velocity=False, return_acceleration=False):
        """ Returns an rao transformed to the body point, pt = (x,y,z)
            raos have shape (6, nBeta, nPeriods)
            body_pts (Np, 3)
            linear_xyz (Np, 3, nBeta, nPeriod)
        """
        X = self.wave_to_motion_tfs
        assert X is not None, "Need to solve the solution first using solve_EOM or solve_stochastic!"
        single_point = True if body_pts.ndim == 1 else False
        npoints = 1 if single_point else len(body_pts)
        linear_xyz = np.zeros((npoints, *X[3:].shape), np.complex)
        if single_point:
            linear_xyz[0] = X[:3] + np.cross(X[3:], body_pts, axisa=0, axisc=0) 
        else:
            for ipt, pt in enumerate(body_pts):
                linear_xyz[ipt] = X[:3] + np.cross(X[3:], pt, axisa=0, axisc=0) 
        jw = 1j*(2*np.pi/self.periods)
        if return_velocity: return linear_xyz*jw
        elif return_acceleration: return linear_xyz*(jw)**2
        else: return linear_xyz

    def _compute_BQE(self, Sxx, fz):
        X = self.wave_to_motion_tfs
        if X is None: 
            self.solve_EOM(amp=1.0) # make an initial (recently changed to facilitate chaining)
            X = self.wave_to_motion_tfs 
        BQE = self._data[Wamit.Properties.LINEARIZED_QUADRATIC_DAMPING_EXTERNAL]
        BQ = self._data[Wamit.Properties.QUADRATIC_DAMPING_EXTERNAL]
        # compute the standard deviation of response for each direction
        for imode, Xi in enumerate(X): # for each response mode and ... 
            for ibeta, beta in enumerate(self.directions): # .. each direction
                # standard deviations of response
                H = interpc(fz, 1/self.periods, np.squeeze(Xi[ibeta]))
                V = np.trapz(Sxx*np.abs(1j*(2*np.pi*fz)*H)**2, fz)**0.5 # Std dev
                BQE[ibeta,imode,imode] = (8/np.pi)**0.5 * V * BQ[imode, imode] # No need to store V = BQE*sqrt(pi/8)
        self._data[Wamit.Properties.LINEARIZED_QUADRATIC_DAMPING_EXTERNAL] = BQE
        return BQE*1.0 # unsure about python references


    def solve_EOM(self, amp=1.0, vstd=None, slender_on=True, print_BSUM=False):
        "Solves the equation of motion"
        import matplotlib.pyplot as plt

        np.set_printoptions(precision=2)
        M = self._data[Wamit.Properties.MASS]
        A = self._data[Wamit.Properties.ADDED_MASS]
        AE = self._data[Wamit.Properties.ADDED_MASS_ZERO_EXTERNAL]
        B = self._data[Wamit.Properties.DAMPING]
        BE = self._data[Wamit.Properties.DAMPING_EXTERNAL]
        C = self._data[Wamit.Properties.STIFFNESS_HYDROGRAV]
        CE = self._data[Wamit.Properties.STIFFNESS_EXTERNAL]
        F = self._data[Wamit.Properties.EXCITE_DIFF]
        X = self._data[Wamit.Properties.RAOS]
        BQE = self._data[Wamit.Properties.LINEARIZED_QUADRATIC_DAMPING_EXTERNAL]

        # Solve for all direction 0, all periods
        X_ = np.zeros(X.shape, np.complex)
        
        for ibeta, beta in enumerate(self.directions):

            #print(f"Heading {beta:.1f}:\nBE={BE}\nBQE={BQE[ibeta]}")
            for iper, period in enumerate(self.periods):
                w = 2*np.pi/period

                # If slender elements are present compute 
                if self._data[Wamit.Properties.SLENDER_ELEMENTS] is not None and slender_on:
                    BV = self._evaluate_slender_element_damping(iper, ibeta, amp, vstd=vstd)
                    FV = self._evaluate_slender_element_wave_excitation(iper, ibeta, amp, vstd=vstd)
                    #print(f"Period, Heading: {period, beta}")
                    #print(f"Viscous damping\n: {BV}")
                    #print(f"Viscous excitation\n: {FV}")
                    #sys.exit()   

                lhs = -w**2*(M+A[:,:,iper]+AE)+1j*w*(B[:,:,iper]+BE+BQE[ibeta])+(C+CE)
                rhs = F[:,ibeta,iper]*1.0

                # TODO: Add viscous wave excitation to rhs (initial check?
                #g = 9.80665
                #z = -22
                #x, y = 0, 0
                #k = w**2/g
                #eta = np.exp(-1j*(k*x*np.cos(beta) + k*y*np.sin(beta))) # phase shift to x
                #u  = np.r_[w*np.cos(beta)*np.exp(k*z)*eta,
                #           w*np.sin(beta)*np.exp(k*z)*eta,
                #           1j*w*np.exp(k*z)*eta, 0, 0, 0]
                #print(rhs)
                #rhs[2] = rhs[2] + BQE[ibeta,2,2]*u[2]
                #rhs[2] = rhs[2] + (BQE[ibeta,2,2]*1j*w*np.exp(k*z))

                if self._data[Wamit.Properties.SLENDER_ELEMENTS] is not None and slender_on:
                    rhs += FV # getting heave excitation effect due to heave pitch damping 
                    lhs += 1j*w*BV

                #x = np.linalg.solve(lhs, rhs)
                X_[:,ibeta,iper] = np.linalg.solve(lhs, rhs)

        # Update internals (RAOs and BQE)
        self._data[Wamit.Properties.RAOS] = X_
        #return X_
        return self # to facilitate chaining

    def solve_stochastic(self, Sxx, fz, verbose_heading=None):
        tol = 1E-7
        first_iteration = True
        factors = np.r_[1,1,1,180/np.pi,180/np.pi,180/np.pi]
        if verbose_heading is not None:
            BQdiag = np.diag(self._data[Wamit.Properties.QUADRATIC_DAMPING_EXTERNAL])
            print(BQdiag)

        while True:
            BQE_old = self._compute_BQE(Sxx, fz) if first_iteration else BQE_new
            #if verbose_heading is not None:
            #    tmp = np.diag(BQE_old[verbose_heading])*(np.pi/8)**0.5
            #    print(np.where(BQdiag>0, tmp/BQdiag, 0)*factors)
            self.solve_EOM() # solve the equations
            BQE_new = self._compute_BQE(Sxx, fz)                
            error = np.sum(np.abs(BQE_old-BQE_new))
            if verbose_heading is not None:
                print(f"Solving Stochastically: Current error {error:.10f}")
                print(np.diag(BQE_old[verbose_heading]))
            if error < tol: break   
            first_iteration = False 
        if verbose_heading is not None:
                tmp = np.diag(BQE_new[verbose_heading])*(np.pi/8)**0.5
                print([tmp[i]/BQdiag[i]*factors[i] if BQdiag[i]!=0 else 0 for i in range(6)])

        return self        


    def _update_retardation_functions(self):
        self._data[Wamit.Properties.RETARDATION_FUNCTIONS] = \
            Wamit.compute_retardation_functions(
                self._data[Wamit.Properties.PERIODS],
                self._data[Wamit.Properties.DAMPING], 
                dt=0.5, tmax=100)

    ###  --------- PROPERTY SECTION ----------  ####

    @property
    def periods(self): return self._data[Wamit.Properties.PERIODS]

    @property
    def directions(self): return self._data[Wamit.Properties.DIRECTIONS]

    @property
    def natural_periods(self): return self._data[Wamit.Properties.NATURAL_PERIODS]

    @property
    def critical_damping(self): return self._data[Wamit.Properties.CRITICAL_DAMPING]

    @property
    def retardation_functions(self): return self._data[Wamit.Properties.RETARDATION_FUNCTIONS]

    @property
    def linear_external_damping(self): return self._data[Wamit.Properties.DAMPING_EXTERNAL]

    @property
    def linearized_quadratic_damping(self): return self._data[Wamit.Properties.LINEARIZED_QUADRATIC_DAMPING_EXTERNAL]

    @property
    def linear_external_stiffness(self): return self._data[Wamit.Properties.STIFFNESS_EXTERNAL]

    @property
    def quadratic_external_damping(self): return self._data[Wamit.Properties.QUADRATIC_DAMPING_EXTERNAL]

    @property
    def external_added_mass(self): return self._data[Wamit.Properties.ADDED_MASS_ZERO_EXTERNAL]

    @property
    def wave_to_motion_tfs(self): return self._data[Wamit.Properties.RAOS]

    @property
    def mass_matrix(self): return self._data[Wamit.Properties.MASS]

    @property
    def body_panels_info(self): return self._data[Wamit.Properties.BODY_PANELS]

    @property
    def body_pressure_tfs(self): return self._data[Wamit.Properties.BODY_PRESSURE_TRANSFER_FUNCTIONS]


    ###  --------- STATIC METHODS ----------  ####

    @staticmethod
    def compute_retardation_functions(periods, damping_matrix, dt=0.5, tmax=100):
        w = 2*np.pi/periods
        wi = np.linspace(1E-12, w.max(), 1000)
        time = np.arange(0,tmax+dt/2,dt)
        rets = np.zeros((6,6,len(time))) # retardation fns 
        for i in range(6):
            for j in range(6):
                bi = np.interp(wi, w, damping_matrix[i,j,:]) 
                rets[i,j]= np.r_[[2/np.pi*np.trapz(bi*np.cos(wi*t), wi) for t in time]]
        return rets

    @staticmethod
    def read_hydrograv_stiffness(filename, ulen=1.0, density=1.025, g=9.80665):
        c = np.zeros((6,6))
        with open(filename) as fin:
            lines = fin.readlines()
            location = find_occurrences("gravitational", lines, specificOccurrenceWanted=1)
            c[2,2:5] = [float(value) for value in lines[location+1].split()[-3:]]
            c[3,3:]  = [float(value) for value in lines[location+2].split()[-3:]]
            c[4,4:]  = [float(value) for value in lines[location+3].split()[-2:]]
            c = c + c.T - np.diag(np.diag(c))
        i, j =  np.indices((6,6))
        k = i//4 + j//4 + 2
        c = c*density*g*ulen**k
        wamit = Record(f"Stiffness properties from {filename}")
        wamit.addattr(stiffness=c)
        return wamit


    @staticmethod
    def read_mass_matrix(filename):
        I, J, mass, damping, stiffness = loadtxt(filename, skip_header=12, transpose=True)
        mass, damping, stiffness = [arr.reshape((6,6)) for arr in [mass, damping, stiffness]]
        get_values = lambda string, lines: np.r_[[float(value) for value in 
            lines[find_occurrences(string, lines, specificOccurrenceWanted=1)].split()[-3:]]]

        with open(filename, "r") as fin:
            lines = fin.readlines()
            vols = get_values("Volumes", lines)
            cob = get_values("Center of Buoyancy", lines)
            cog = get_values("Center of Gravity", lines)

        wamit = Record(f"Mass properties from {filename}")
        wamit.addattr(mass=mass, damping=damping, stiffness=stiffness, cob=cob, cog=cog, vols=vols)
        return wamit

    @staticmethod
    def read_added_mass_and_damping(filename, ulen=1.0, density=1.025, frequency_ascending=False):
        """ Returns the dimensionalized added mass
        """
        periods = np.loadtxt(filename, usecols=(0,))
        periods = np.r_[sorted(set(periods))]
        nperiods = len(periods) - np.sum(periods <= 0)
        periods = periods[periods > 0]
        a0, ainf = np.zeros((2,6,6))
        added_mass = np.zeros((6, 6, nperiods))
        damping = np.zeros((6, 6, nperiods))
        with open(filename, 'r') as fin:
            for line in fin.readlines():
                vars = [float(var) for var in line.split()]
                # note that for periods 0 and -1 bij is dummy        
                (period, i, j, aij), bij = vars[:4], vars[-1] 
                i, j = int(i), int(j)
                k = i//4 + j//4 + 3
                aij = aij*density*ulen**k
                if period == 0:
                    a0[i-1, j-1] = aij
                elif period < 0:
                    ainf[i-1, j-1] = aij
                else:
                    bij = bij*density*ulen**k*(2*np.pi/period)
                    added_mass[i-1, j-1,
                               np.argmin(abs(periods-period))] = aij
                    damping[i-1, j-1,
                               np.argmin(abs(periods-period))] = bij
        if frequency_ascending:
            periods = periods[::-1]
            added_mass = added_mass[:,:,::-1]
            damping = damping[:,:,::-1]
        wamit = Record("periods from %s" % filename)
        wamit.addattr(periods=periods, added_mass=added_mass, damping=damping, ainf=ainf, a0=a0)
        return wamit

    @staticmethod
    def read_excitation_forces(filename, ulen=1.0, density=1.025, g=9.80665, frequency_ascending=False):
        # Format PER BETA I Mod(Xi) Pha(Xi) Re(Xi) Im(Xi)
        periods, directions, dofs, amp, phase, re, im =  loadtxt(filename, transpose=True)
        periods = np.r_[sorted(set(periods))]
        directions = np.r_[sorted(set(directions))]
        value = (re + 1j*im)*density*g*ulen**(1+dofs//4)
        dofs = sorted(set(dofs))
        value = np.einsum('lkj->jkl', value.reshape((len(periods), len(directions), len(dofs))))
        if frequency_ascending:
            periods = periods[::-1]
            value = value[:,:,::-1]
        tfs = Record(f'Excitation forces from {"Haskind Relations" if filename.endswith("2") else "Diffraction Potential"}')
        tfs.addattr(value=value, directions=directions, periods=periods)

        #print(filename)
        #if filename == "wamit.3":
        #    for iper, period in enumerate(periods):
        #        if period == 10:
        #            print(value[2,0,iper], density*g)
        #
        #    sys.exit()
        return tfs

    @staticmethod
    def read_raos(filename="wamit_.4", ulen=1.0, nbodies=1, frequency_ascending=False):
        """ Reads a wamit.4 file - nbodies not implemented"""
        periods, directions, dofs, amp, phase, re, im =  loadtxt(filename, transpose=True)
        periods = np.r_[sorted(set(periods))]
        directions = np.r_[sorted(set(directions))]
        value = (re + 1j*im)/ulen**(dofs//4)
        value = np.einsum('lkj->jkl', value.reshape((len(periods), len(directions), 6)))
        if frequency_ascending:
            periods = periods[::-1]
            value = value[:,:,::-1]
        tfs = Record(f'Wave to motion transfer function')
        tfs.addattr(value=value, directions=directions, periods=periods)
        return tfs

    @staticmethod
    def compute_wavenumber(w, h, g = 9.80665):
        """ Returns the wave number satisfying the first order linear dispersion relation w**2 = gktanh(kh)
        """
        from scipy.optimize import root_scalar
        # To relax requirement on input allow for both iterable and non-iterable inputs
        made_iterable = False
        if not hasattr(w,'__iter__'): 
            w = np.ones(1)*w # make iterable
            made_iterable = True

        k = np.zeros(len(w))
        for i, wi in enumerate(w):
            sol = root_scalar(lambda k_, h, w: w**2 - g*k_*np.tanh(k_*h), 
                                args=(h, wi), method='brentq', bracket=[0, 10])
            if not sol.converged: print(f"Period {2*np.pi/wi} is not converged.")
            k[i] = sol.root

        return k[0] if made_iterable else k

    @staticmethod
    def compute_incident_wave(periods, directions, fpts, depth=1000, g=9.80665, 
            return_velocity=False, return_acceleration=False):
        """ 
        Computes the incident wave (origin) to wave (field point) transfer function
        TODO: Investigate computation of shallow water depth dependency term
        WAMIT quantities are phased with respect to the incident wave elevation at X=Y=0 
        """
        from scipy.optimize import minimize
        import warnings
        warnings.filterwarnings("ignore")
        omega = 2*np.pi/periods
        k = Wamit.compute_wavenumber(omega, depth, g)
        kx = np.zeros((len(fpts), len(directions), len(periods)), np.complex)
        u = np.zeros((len(fpts), 3, len(directions), len(periods)), np.complex)
        # Evaluate the wave-to-wave elevation tf at the field points
        jw = 1j*omega
        factor = jw if return_acceleration else 1 if return_velocity else 1/jw # for position

        for ib, beta in enumerate(directions*np.pi/180):
            for ip, (x,y,z) in enumerate(fpts):
                kx[ip, ib] = np.exp(-1j*(k*x*np.cos(beta) + k*y*np.sin(beta))) # phase shift for xy
                #u[ip, ib] = np.cos(beta)*omega*np.cosh(k*(z+depth))/np.sinh(k*depth)*kx[ip, ib]
                #v[ip, ib] = np.sin(beta)*omega*np.cosh(k*(z+depth))/np.sinh(k*depth)*kx[ip, ib]
                #w[ip, ib] = 1j*omega*np.sinh(k*(z+depth))/np.sinh(k*depth)*kx[ip, ib]
                u[ip, 0, ib] = np.cos(beta)*omega*np.exp(k*z)*kx[ip, ib]*factor # horizontal is in phase with elevation
                u[ip, 1, ib] = np.sin(beta)*omega*np.exp(k*z)*kx[ip, ib]*factor # horizontal is in phase with elevation
                u[ip, 2, ib] = 1j*omega*np.exp(k*z)*kx[ip, ib]*factor           # vertical is -sin
        return u

    @staticmethod
    def read_body_pressure_tfs(filename="wamit.5p", nbodies=1):
        """ Reads a wamit.5p file - nbodies not implemented  - assumes INUMOPT5 = 1"""
        data = open(filename, "r").readlines()
        data = [list(map(float, line.split())) for line in data]
        data_rad = np.array([line for line in data if len(line)==15]).T
        data_dif = np.array([line for line in data if len(line)==6]).T
        wave_periods, directions, quadrant, pid = data_dif[:4]
        wave_periods = np.sort(np.r_[list(set(wave_periods))])
        directions = np.sort(np.r_[list(set(directions))])
        nf, nd = len(wave_periods), len(directions)
        npanels = int(max(pid)*max(quadrant))
        re_rad, im_rad = data_rad[3::2], data_rad[4::2]
        re_dif, im_dif = data_dif[4], data_dif[5]
        h_rad = re_rad + 1j * im_rad
        h_dif = re_dif + 1j * im_dif
        h_rad = np.einsum('ijk->ijk', h_rad.reshape((6, nf, npanels))) # leave unchanged
        h_dif = np.einsum('kjl->jkl', h_dif.reshape((nf, nd, npanels)))

        # reverse periods to suit SIMO and aid integrations using w
        wave_periods = wave_periods[::-1]
        h_dif = np.flip(h_dif, axis=1) # flip frequency axis (index 1 after einsum)
        h_rad = np.flip(h_rad, axis=1) # flip frequency axis

        pres = Record("Pressure on body from %s" % filename)
        pres.directions = directions
        pres.frequencies = 2 * np.pi / wave_periods
        pres.periods = wave_periods
        pres.h_rad = h_rad
        pres.h_dif = h_dif

        for ibody in range(nbodies):
            for idof, name in enumerate("surge sway heave roll pitch yaw".split()):
                pres.setattr(name, h_rad[idof])

        return pres

    @staticmethod
    def read_wamit_normals(filename="wamit.pnl", nbodies=1):
        xc, yc, zc, area, nx, ny, nz, rnx, rny, rnz = np.loadtxt(filename, usecols=(2, 3, 4, 5, 6, 7, 8, 9, 10, 11)).T
        geom = Record("Areas and normals from %s" % filename)
        geom.area = area
        geom.xc  = xc
        geom.yc  = yc
        geom.zc  = zc
        geom.nx  = nx 
        geom.ny  = ny 
        geom.nz  = nz 
        geom.rnx = rnx
        geom.rny = rny
        geom.rnz = rnz
        geom.npanels = len(xc)
        return geom
    

############ TO BE DEPRECATED BELOW THIS LINE #############


def read_simo_raos(filename, bodyname=None):
    """ If the file contains more than one body, then you need to identify the body"""

    # Create record for storage
    rao = Record("rao from %s" % filename)

    # alias the data appropriately
    lines = open(filename, 'r').readlines()

    # strip whitespace
    lines = list([line for line in lines
                  if not line.strip().startswith("'")])

    # find occurances of wave drift (assume none in text)
    loc = [i+1 for i, l in enumerate(lines)
           if 'motion' in l.lower() and 'transfer' in l.lower()]

    # read number of frequencies
    nd, nf = map(int, lines[loc[0]+2].split()[:2])

    # read frequencies and directions
    rao.directions = \
        np.r_[list(map(float, [l.split()[1]
                               for l in lines[loc[1]:loc[1]+nd]]))]
    rao.frequencies = \
        np.r_[list(map(float, [l.split()[1]
                               for l in lines[loc[2]:loc[2]+nf]]))]

    # read RAOs forces
    for i in loc[3:]:
        data = np.r_[[float(l.split()[2]) *
                      np.exp(1j*float(l.split()[3])*np.pi/180.)
                      for l in lines[i:i+nd*nf]]]
        data = data.reshape((nd, nf))
        if 'surge' in lines[i-1].lower():
            rao.surge = data
        elif 'sway' in lines[i-1].lower():
            rao.sway = data
        elif 'heave' in lines[i-1].lower():
            rao.heave = data
        elif 'roll' in lines[i-1].lower():
            rao.roll = data
        elif 'pitch' in lines[i-1].lower():
            rao.pitch = data
        elif 'yaw' in lines[i-1].lower():
            rao.yaw = data

    return rao


def read_wamit_raos(filename="wamit_.4", nbodies=1):
    """ Reads a wamit.4 file - nbodies not implemented"""
    wave_periods, directions, dof, amp, phase, re, im = np.loadtxt(filename).T
    wave_periods = np.sort(np.r_[list(set(wave_periods))])
    directions = np.sort(np.r_[list(set(directions))])
    nf, nd = len(wave_periods), len(directions)
    re = np.einsum('lkj->jkl', re.reshape((nf, nd, 6)))
    im = np.einsum('lkj->jkl', im.reshape((nf, nd, 6)))
    h = re + 1j * im

    # reverse periods to suit SIMO
    wave_periods = wave_periods[::-1]
    h = h[:, :, ::-1]

    rao = Record("RAOs from %s" % filename)
    rao.directions = directions
    rao.frequencies = 2 * np.pi / wave_periods
    rao.periods = wave_periods
    rao.h = h

    for ibody in range(nbodies):
        for idof, name in enumerate("surge sway heave roll pitch yaw".split()):
            rao.setattr(name, h[idof, :, :].reshape((nd, nf)))

    return rao


def read_wamit_mean_drift(filename="wamit_.8", nbodies=1, rho=1.025, g=9.80665, momentum=True):
    """ Reads a wamit.8 file - nbodies not implemented"""
    wave_periods, directions, direction2, dof, amp, phase, re, im = np.loadtxt(filename).T
    wave_periods = np.sort(np.r_[list(set(wave_periods))])
    directions = np.sort(np.r_[list(set(directions))])
    nf, nd = len(wave_periods), len(directions)
    re = np.einsum('lkj->jkl', re.reshape((nf, nd, 3 if momentum else 9)))
    im = np.einsum('lkj->jkl', im.reshape((nf, nd, 3 if momentum else 9)))
    h = re + 1j * im
    # reverse periods to suit SIMO
    wave_periods = wave_periods[::-1]
    h = h[:, :, ::-1]*rho*g
    # create record for export
    wdrift = Record("Wave drift (momentum) from %s" % filename)
    wdrift.directions = directions
    wdrift.frequencies = 2 * np.pi / wave_periods
    wdrift.periods = wave_periods
    wdrift.h = h
    # the last 3 in the case of pressure are relative the moving origin as opposed to mean position
    names = "surge sway yaw" if momentum else "surge sway heave roll pitch yaw roll2 pitch2 yaw2"
    for ibody in range(nbodies):
        for idof, name in enumerate(names.split()):
            wdrift.setattr(name, h[idof, :, :].reshape((nd, nf)))
    return wdrift



def read_wave_elevation_tfs(filename="wamit_.6p"):
    wave_periods, directions, fpts, amp, phase, re, im = np.loadtxt(filename).T
    wave_periods = np.sort(np.r_[list(set(wave_periods))])
    directions = np.sort(np.r_[list(set(directions))])
    fpts = np.sort(np.r_[list(set(fpts))])
    nf, nd, nfpts = len(wave_periods), len(directions), len(fpts)
    re = np.einsum('lkj->jkl', re.reshape((nf, nd, nfpts)))
    im = np.einsum('lkj->jkl', im.reshape((nf, nd, nfpts)))
    h = re + 1j * im

    # reverse periods to suit SIMO
    wave_periods = wave_periods[::-1]
    h = h[:, :, ::-1]

    wave = Record("Wave elevation from %s" % filename)
    wave.directions = directions
    wave.frequencies = 2 * np.pi / wave_periods
    wave.periods = wave_periods
    wave.elev = h

    return wave


def export_raos_to_simo(raos, body_name="body"):
    from ..administration.tools import sed
    sysfile = "sys-raos.dat"
    infile = sysfile + "0"
    outfile = sysfile
    beta = raos.directions
    w = raos.frequencies

    xxxdirs = ''.join(['\n%d  %.1f' % (i+1, d) for i, d in enumerate(beta)])
    xxxfreqs = ''.join(["\n %6d   %f" % (i+1, f) for i, f in enumerate(w)])
    sed(infile, {"xxxndirs": len(beta), "xxxnfreq": len(w)}, outfile)

    infile = outfile
    sed(infile, {"xxxdirs": xxxdirs, "xxxfreqs": xxxfreqs}, outfile)
    for idof, name in enumerate("surge sway heave roll pitch yaw".split()):
        L = []
        for ihead in range(len(beta)):
            for ifreq in range(len(w)):
                L.append('\n %6d %6d %12.3e %12.3e' % (
                    ihead + 1, ifreq + 1,
                    np.abs(raos.h[idof, ihead, ifreq]),
                    np.angle(raos.h[idof, ihead, ifreq])*180./np.pi))

        sed(infile, {"xxx%sRAOs" % name: ''.join(L)}, outfile)
        sed(infile, {"xxxBODY": body_name}, outfile)
    try:
        os.remove("sys-%s.dat" % body_name)
    except:
        pass
    os.rename("sys-raos.dat", "sys-%s.dat" % body_name)


def read_added_mass(filename="wamit_.1", ulen=1.0, density=1.025):
    """ Returns the dimensionalized added mass
    """
    data = np.loadtxt(filename, usecols=(0, 1, 2, 3))
    period, I, J, added_mass = data.T
    # period = set(period)  # tmp to remove 0 and -1
    # period.remove(0)   # tmp to remove 0 and -1
    # period.remove(-1)  # tmp to remove 0 and -1
    periods = np.r_[sorted(set(period))]
    added_mass = np.zeros((6, 6, len(periods)))
    for row in data:
        period, i, j, aij = row
        if period in [-1, 0]:
            continue
        else:
            added_mass[int(i-1), int(j-1),
                       np.argmin(abs(periods-period))] = aij*density

    wamit = Record("periods from %s" % filename)
    wamit.periods = periods
    wamit.added_mass = added_mass
    return wamit


def read_added_mass_and_damping(filename="wamit_.1", ulen=1.0, density=1.025):
    """ Returns the dimensionalized added mass
    """
    periods = np.loadtxt(filename, usecols=(0,))
    periods = np.r_[sorted(set(periods))]
    nperiods = len(periods) - np.sum(periods <= 0)
    periods = periods[periods > 0]
    a0, ainf = np.zeros((2,6,6))
    added_mass = np.zeros((6, 6, nperiods))
    damping = np.zeros((6, 6, nperiods))
    with open(filename, 'r') as fin:
        for line in fin.readlines():
            vars = [float(var) for var in line.split()]
            # note that for periods 0 and -1 bij is dummy        
            (period, i, j, aij), bij = vars[:4], vars[-1] 
            i, j = int(i), int(j)
            k = i//4 + j//4 + 3
            aij = aij*density*ulen**k
            bij = bij*density*ulen**k
            if period == 0:
                a0[i-1, j-1] = aij
            elif period < 0:
                ainf[i-1, j-1] = aij
            else:
                added_mass[i-1, j-1,
                           np.argmin(abs(periods-period))] = aij
                damping[i-1, j-1,
                           np.argmin(abs(periods-period))] = bij

    wamit = Record("periods from %s" % filename)
    wamit.addattr(periods=periods, added_mass=added_mass, damping=damping, ainf=ainf, a0=a0)
    return wamit


def read_motion_tfs_not_used(filename="wamit_.4"):
    # laod and unpack data
    data = np.loadtxt(filename)
    period, direction, dof, amp, phase, re, im = data.T
    # reshape arrays
    periods = sorted(set(period))
    directions = sorted(set(direction))
    dofs = sorted(set(dof))
    mats = [period, direction, dof, amp, phase, re, im]  # pack
    for i in range(len(mats)):
        mats[i] = mats[i].reshape(len(periods), len(directions), len(dofs))
    period, direction, dof, amp, phase, re, im = mats  # unpack
    # package for returning
    tfs = Record("motion_tfs")
    tfs.addattr(amp=amp, phase=phase, re=re, im=im,
                dirs=direction, pers=period,
                directions=directions, periods=periods)
    # if direction and periods not order use einsum to reorder
    return tfs


def read_wave_elevation_tfs_not_used(filename="wamit_.6p"):
    data = np.loadtxt(filename)
    period, direction, fpt, amp, phase, re, im = data.T
    periods = sorted(set(period))
    directions = sorted(set(direction))
    fpts = sorted(set(fpt))
    mats = [period, direction, fpt, amp, phase, re, im]  # pack
    for i in range(len(mats)):
        mats[i] = mats[i].reshape(len(periods), len(directions), len(fpts))
    period, direction, fpt, amp, phase, re, im = mats  # unpack
    # package for returning
    tfs = Record("wave_elevation")
    tfs.addattr(amp=amp, phase=phase, re=re, im=im,
                dirs=direction, pers=period,
                directions=directions, periods=periods)
    return tfs



r"""A python module for spectral analysis

    Functions:
    
    specpack.autospec
    specpack.compspec
    specpack.crosspec
    specpack.gaussiansmooth
    specpack.plotrao
    specpack.plotspec
    specpack.rao
    specpack.specdata
    specpack.specmoment

    """


import numpy as np
import scipy.signal as sg
import matplotlib.pyplot as plt
import math

###########################################################
###########################################################

def phase_autospectrum(S, seed=None):
    """ Energy auto-spectrum S(f) 
    For wave elevation the units are m*m*s
    """
    np.random.seed(seed)
    phases = np.random.rand(len(S))*2*np.pi
    return S*np.exp(1j*phases)

###########################################################
###########################################################

def autospec(x,t,smooth=False,bw=0.001,welch=False,**kwargs):
    r"""specpack.autospec
        -----------------

        Calculates auto-spectrum of the time series x
        Returns the one-sided auto-spectrum
        
        Option to smooth:
            1.  welch = False : applies a Gaussian smoothing in the pseudo time domain
                smooth = True
                bw = spectral bandwidth of Gaussian

            2.  welch = True : uses scipy.signal welch method. Accepted kwargs are 
                nseg - number of segments (defaults to 30)
                window - defaults to a Gaussian with std 0.25 the segment length


        -----------------

        Input:   x - time series 
                 t - time vector
                 bw - smoothing parameter

        Output:  G_xx - auto-spectrum of x 
                 fz  - corresponding frequency vector

        -----------------"""

    assert(len(x) == len(t)), "time and signal array must be equal"

    # Time and frequency vector characteristics
    dt = t[1]-t[0]
    Nt = len(t)   

    if welch: # use scipy's welch method (removes mean)
        nseg = kwargs["nseg"] if "nseg" in kwargs else 30
        nperseg = len(x) // nseg
        window = kwargs["window"] if "window" in kwargs else sg.windows.gaussian(nperseg,std=nperseg/4)
        n = np.log(nperseg)//np.log(2) + 3  # unsure about this nfft = 2**16 works with test case
        fz, Sxx = sg.welch(x, fs = 1/dt, nperseg = nperseg, window=window, nfft=2**n) 
        return Sxx, fz

    # Remove mean 
    x = x-np.mean(x)

    # zero padding !
    Nn = 2**int(math.ceil(math.log(Nt,2)))
    df = 1/Nn/dt  # zero padded frequency step size  
    fz = np.arange(0, Nn//2)*df

    # Calculate Fourier transform
    X = np.fft.fft(x,Nn)

    # Calculate auto-spectrum
    # The Nt is to normalize the spectrum so it is not a function of r.length 
    # The *2 is to keep the same energy (as we drop half the spectrum)
    # The dt is to convert the power to an energy spectrum
    G_xx =np.abs(X)**2/Nt * 2*dt

    return (gaussiansmooth(G_xx[:Nn//2],fz,bw).real, fz) if smooth else (G_xx[:Nn//2].real,fz) 
    
###########################################################
###########################################################

def crosspec(x, y, t, smooth=False, bw=0.001, welch=False, **kwargs):
    r"""specpack.crosspec
        -----------------
        
        Calculates cross-spectrum of two time series

        -----------------

        Input:     x - Time series 1
                   y - Time series 2
                   t - time vector

        Option to smooth:
            1.  welch = False : applies a Gaussian smoothing in the pseudo time domain
                smooth = True
                bw = spectral bandwidth of Gaussian

            2.  welch = True : uses scipy.signal welch method. Accepted kwargs are 
                nseg - number of segments (defaults to 30)
                window - defaults to a Gaussian with std 0.25 the segment length

        Output:    G_xy - Cross-spectrum of x and y
                   fz  - corresponding frequency vector
                   
        -----------------"""

    assert(len(x) == len(y) == len(t)), "time and signal arrays must be equal length"

    # Time and frequency vector characteristics
    dt = t[1]-t[0]
    Nt = len(t)

    if welch: # use scipy's welch method (removes mean)
        nseg = kwargs["nseg"] if "nseg" in kwargs else 30
        nperseg = len(x) // nseg
        window = kwargs["window"] if "window" in kwargs else sg.windows.gaussian(nperseg,std=nperseg/4)
        n = np.log(nperseg)//np.log(2) + 3  # unsure about this nfft = 2**16 works with test case
        fz, Sxy = sg.csd(x, y, fs = 1/dt, nperseg = nperseg, window=window, nfft=2**n) 
        return Sxy, fz


    # Remove mean
    x = x-x.mean()
    y = y-y.mean()

    # zero padding !
    Nn = 2**int(math.ceil(math.log(Nt,2)))
    df = 1/Nn/dt  # zero padded frequency step size  
    fz = np.arange(0, Nn//2)*df

    # Calculate Fourier transforms
    X = np.fft.fft(x,Nn)
    Y = np.fft.fft(y,Nn)

    # Calculate cross-spectrum (reference autospectra)
    G_xy = X.conjugate()*Y/Nt * 2*dt

    return (gaussiansmooth(G_xy[:Nn//2],fz,bw), fz) if smooth else (G_xy[:Nn//2],fz)


###########################################################
###########################################################

def coherence(x, y, t, smooth=False, bw=0.001, welch=False, spectra=False, **kwargs):
    r"""specpack.coherencec
        -----------------
        
        Calculates coherence of two time series x(t) and y(t)

        -----------------

        Input:     x - Time series 1
                   y - Time series 2
                   t - time vector

        Option to smooth:
            1.  welch = False : applies a Gaussian smoothing in the pseudo time domain
                smooth = True
                bw = spectral bandwidth of Gaussian

            2.  welch = True : uses scipy.signal welch method. Accepted kwargs are 
                nseg - number of segments (defaults to 30)
                window - defaults to a Gaussian with std 0.25 the segment length

        Output:     C_xy - coherence of x and y
                    fz   - corresponding frequency vector

                If spectra = True, also returns the additional fields

                    G_xy - cross-spectrum of x and y
                    G_xx - auto-spectrum of x
                    G_yy - auto-spectrum of y

        -----------------"""
    G_xy, fz = crosspec(x, y, t, smooth=smooth, bw=bw, welch=welch, **kwargs)
    G_xx, fz = autospec(x, t, smooth=smooth, bw=bw, welch=welch, **kwargs)
    G_yy, fz = autospec(y, t, smooth=smooth, bw=bw, welch=welch, **kwargs)
    C_xy = np.where(G_xx*G_yy>0, np.abs(G_xy)**2/(G_xx*G_yy), 0)
    return (C_xy, fz, G_xy, G_xx, G_yy) if spectra else (C_xy, fz)

###########################################################
###########################################################


def rao(x, y, t, phase=True, smooth=False, bw=0.001, welch=False, spectra=False, **kwargs):
    r"""specpack.rao
        -----------------------
       
        Calculates RAO
 
        Uses:    specpack.autospec
                 specpack.crosspec
                 specpack.gaussiansmooth

        -----------------------

        Input:  x - displacement time series
                y - wave time series
                t - corresponding time vector

        Option to smooth:
            1.  welch = False : applies a Gaussian smoothing in the pseudo time domain
                smooth = True
                bw = spectral bandwidth of Gaussian

            2.  welch = True : uses scipy.signal welch method. Accepted kwargs are 
                nseg - number of segments (defaults to 30)
                window - defaults to a Gaussian with std 0.25 the segment length

        Output: H1 - linear x to y transfer function
                fz  - corresponding frequency vector

            If phase = True, returns the complex transfer function H1
            If phase = False, returns the absolute of H1 (aka RAO).

            If spectra = True, also returns the additional fields

                G_xy - cross-spectrum of x and y
                G_xx - auto-spectrum of x
                G_yy - auto-spectrum of y
                C_xy - correlation coefficient of x and y

        -----------------------"""

    C_xy, fz, G_xy, G_xx, G_yy = coherence(x, y, t, smooth=smooth, bw=bw, welch=welch, spectra=True, **kwargs)
    H1 = np.where((C_xy > 0.7) & (G_yy/G_yy.max() > 0.05), G_xy/G_yy if phase else np.sqrt(G_xx / G_yy), 0)
    return (H1, fz, G_xy, G_xx, G_yy, C_xy) if spectra else (H1, fz)


###########################################################
###########################################################



def gaussiansmooth(S,w,s=0.001):
    r"""specpack.gaussiansmooth
        -----------------------
       
        Spectral smoothing by convolving the raw spectrum 
        with a Gaussian window in frequency domain.
        Actual computation performed in "time"-domain 
        to avoid the convolution integral.

        Steps:
        1) Calculate the Fourier transform $\hat{S}(t)$ of the 
           raw spectrum $S(\omega)$
        2) Multiply $\hat{S}(t)$ by $\hat{G}(t)=\exp{-\omega^2 t^2/2}$
        3) The smoothed spectrum is the inverse Fourier transform of
           $\hat{S}(t)\hat{G}(t)$

        -----------------------

        Input:  S  - raw spectrum
                w  - corresponding frequency vector [Hz]
                s  - stdev of Gaussian smoothing function [Hz]

        Output: Ss - Smoothed spectrum

        -----------------------"""

    # Test if raw spectrum is real of complex
    complexInputFlag = 1 if any(S.imag!=0) else 0

    # Frequency vector characteristics
    Nw = len(w)
    dw = w[1]-w[0]
    
    # Add zeros to remove edge effects (FFT/periodicity)
    Nzeros = 2*math.floor(float(s)/dw)
    S2 = np.zeros([2*Nzeros+len(S)], np.complex) #float)
    S2[Nzeros:Nzeros+len(S)] = S
    
    if Nw%2!=0: S2=S2[0:-1]
    N2 = len(S2)
    
    # Define "time" vector
    dt = 2.*np.pi/N2/dw
    t = np.arange(-(N2/2.)*dt,((N2/2.)-1)*dt+dt,dt)
    
    # Gaussian window in time domain
    G = np.exp(-(s**2)*(t**2)/2.)
   
    # Smoothing
    Shat = np.fft.fftshift(np.fft.fft(S2))
    Ss   = np.fft.ifft(np.fft.ifftshift(Shat*G))

    # Remove imaginary residual only if input was real
    if complexInputFlag==1:
        SS = Ss.real
    
    return Ss[Nzeros:Nzeros+len(S)]



###########################################################
###########################################################




def plotspec(fz,Ss,xmax,dof,dim,showflag=0,filename=0,xmin=0,title=''):

    """specpack.plotspec
       -----------------

       Plotting of auto-spectra with scaling as in Plotit

       -----------------

       Input:   fz       - frequency vector [Hz]
                Ss       - auto-spectrum
                xmax     - max frequency in plot
                dof      - name of dof
                dim      - dimension of dof
                showflag - if 1, then figure is shown
                filename - if 0, then nothing is done
                           otherwise, figure is saved as filename.png
                xmin     - min frequency in plot, default is 0
                title    - plot title, default is ''

       Output:  Optional show() of figure
                Optional savefig() of figure

       -----------------"""
    
    plt.figure()
    plt.clf()
    plt.plot(fz,Ss,linewidth=2.0)
    ymax = max(1.05*Ss)
    plt.axis([xmin,xmax,0,ymax])
    plt.xlabel('Frequency [Hz]')
    plt.ylabel('%s auto-spectrum [%s$^2$s]' % (dof,dim))
    plt.grid()
    plt.title(title)
    if filename!=0:
        plt.savefig(filename+'_autospec')
    if showflag==1:
        plt.show()
    

    plt.close()



###########################################################
###########################################################




def plotRAO(fz,Ss,xmax,label,showflag,filename,axisflag=0,timeflag=0,xmin=0,title=''):

    """specpack.plotRAO
       -----------------

       Plotting of RAO
       
       vak 06.05.2011

       -----------------

       Input:   fz       - frequency vector [Hz]
                RAO      - RAO to be plotted
                xmax     - max value on x-axis
                showflag - if 1, then figure is shown
                filename - if 0, then nothing is done
                           otherwise, figure is saved as filename_RAO.png
                axisflag - if 0, do not plot RAO for frequencies 
                                 lower than 0.025 Hz
                           if 1, plot RAO for low frequencies as well
                timeflag - if 0, plot against frequency
                           if 1, plot against time
                           default is 0
                xmin     - min frequency in plot, default is 0
                title    - plot title, default is ''
                           

       Output:  Optional show() of figure
                Optional savefig() of figure

       -----------------"""
    
    plt.figure()
    plt.clf()

    if axisflag==0:
        minfreq = 0.025    
        ind_low = plt.find(fz>minfreq)[0]
    elif axisflag==1:
        ind_low = 1

    ind_top = len(Ss)
    if len(plt.find(fz<xmax)): ind_top = plt.find(fz<xmax)[-1]


    if timeflag==0:
        plt.plot(fz[ind_low:ind_top],Ss[ind_low:ind_top],linewidth=2.0)
        plt.xlabel('Frequency [Hz]')
    elif timeflag==1:
        plt.plot(1/fz[ind_low:ind_top],Ss[ind_low:ind_top],linewidth=2.0)
        plt.xlabel('Period [s]')

    ymax = max(1.05*Ss[ind_low:ind_top])
    plt.axis([xmin,xmax,0,ymax])
    plt.ylabel(label)
    plt.title(title)
    plt.grid()    
    if filename!=0:
        plt.savefig(filename+'_RAO')
    if showflag==1:
        plt.show()

    plt.close()

###########################################################
###########################################################


def plotspec(fz,Ss,xmax,label,showflag,filename,axisflag=0,timeflag=0,xmin=0,title=''):

    """specpack.plotRAO
       -----------------

       Plotting of RAO
       
       vak 06.05.2011

        Modifications:
        tek 13.12.2011 - single-sided returned
       -----------------

       Input:   fz       - frequency vector [Hz]
                spec      - spec to be plotted
                xmax     - max value on x-axis
                showflag - if 1, then figure is shown
                filename - if 0, then nothing is done
                           otherwise, figure is saved as filename_spec.png
                axisflag - if 0, do not plot spec for frequencies 
                                 lower than 0.025 Hz
                           if 1, plot spec for low frequencies as well
                timeflag - if 0, plot against frequency
                           if 1, plot against time
                           default is 0
                xmin     - min frequency in plot, default is 0
                title    - plot title, default is ''
                           

       Output:  Optional show() of figure
                Optional savefig() of figure

       -----------------"""
    
    plt.figure()
    plt.clf()

    if axisflag==0:
        minfreq = 0.025    
        ind_low = plt.find(fz>minfreq)[0]
    elif axisflag==1:
        ind_low = 1

    ind_top = len(Ss)
    if len(plt.find(fz<xmax)): ind_top = plt.find(fz<xmax)[-1]


    if timeflag==0:
        plt.plot(fz[ind_low:ind_top],Ss[ind_low:ind_top],linewidth=2.0)
        plt.xlabel('Frequency [Hz]')
    elif timeflag==1:
        plt.plot(1/fz[ind_low:ind_top],Ss[ind_low:ind_top],linewidth=2.0)
        plt.xlabel('Period [s]')

    ymax = max(1.05*Ss[ind_low:ind_top])
    plt.axis([xmin,xmax,0,ymax])
    plt.ylabel(label)
    plt.title(title)
    plt.grid()    
    if filename!=0:
        plt.savefig(filename+'_spec')
    if showflag==1:
        plt.show()

    plt.close()



###########################################################
###########################################################

def specdata(x,t,s=0.001):

    """specpack.specdata
       -------------------

       Summary of spectral data
       
       Uses:    specpack.autospec
                specpack.gaussiansmooth
                specpack.specmoment

       -------------------

       Input:  x - displacement time series
               t - corresponding time vector
               s - smoothing parameter

       Output: M0      - 0th spectral moment 
               M1      - 1st spectral moment 
               M2      - 2nd spectral moment 
               M4      - 4th spectral moment 
               T1      - energy average period
               T2      - average zero crossing period
               peakper - spectral peak period
               H       - significant value


       -----------------"""
    # Spectral moments
    M0 = specmoment(x,t,0,s)
    M1 = specmoment(x,t,1,s)
    M2 = specmoment(x,t,2,s)
    M4 = specmoment(x,t,4,s)

    # Periods
    #T1 = 2*np.pi*M0/M1 
    #T2 = 2*np.pi*np.sqrt(M0/M2)

    T1 = M0/M1 
    T2 = np.sqrt(M0/M2)

    # Spectral peak period
    [asp,fz] = autospec(x,t)
    asp_s = gaussiansmooth(asp,fz,s)
    peakper = float(1./fz[np.nonzero(asp_s==asp_s.max())])

    # Significant value
    H  = 4*np.sqrt(M0)

    return M0,M1,M2,M4,T1,T2,peakper,H

###########################################################
###########################################################

def specmoment(x,t,n,s):
    
    """specpack.specmoment
       -------------------

       Calculation of spectral moments based on a smoothed spectrum
       
       Uses:    specpack.autospec
                specpack.gaussiansmooth

       -------------------

       Input:  x - displacement time series
               t - corresponding time vector
               n - order of moment
               s - smoothing parameter

       Output: M - nth spectral moment 

       -----------------"""
    
    [asp,fz] = autospec(x,t) # single-sided!

    asp_s = gaussiansmooth(asp,fz,s)
    
    # Use trapezoidal quadrature from numpy 

    N = min(len(fz), len(asp_s))
    M = np.trapz(fz[:N]**n*asp_s[:N],fz[:N])

    return M


###########################################################
###########################################################



def compspec(fz1,Ss1,fz2,Ss2,fw,Sw,lab1,lab2,xmax,dof,dim,
            showflag=0,filename=0,legend=None,
            xmin=0,title='',fz_WADAM=None, Ss_WADAM=None, **kwargs):

    """specpack.compspec
       -----------------

       Plotting of two auto-spectra with scaling as in Plotit
       
       vak 06.05.2011

        Modifications:
        tek 13.12.2011 - single-sided returned
       -----------------

       Input:   fz1      - frequency vector [Hz] of spectrum 1
                Ss1      - auto-spectrum 1
                fz2      - frequency vector [Hz] of spectrum 2
                Ss2      - auto-spectrum 2
                xmax     - max frequency in plot
                dof      - name of dof
                dim      - dimension of dof
                showflag - if 1, then figure is shown
                filename - if 0, then nothing is done
                           otherwise, figure is saved as filename.png
                legend   - default is None
                xmin     - min frequency in plot, default is 0
                title    - plot title, default is ''

       Output:  Optional show() of figure
                Optional savefig() of figure

       -----------------"""
    
    plt.figure()
    plt.clf()
    p1 = plt.plot(fz1,Ss1,linewidth=2.0, label=lab1)
    p2 = plt.plot(fz2,Ss2,linewidth=2.0, label=lab2)
    ymax = max(max(1.05*Ss1),max(1.05*Ss2))
    plt.axis([xmin,xmax,0,ymax])
    plt.xlabel('Frequency [Hz]')
    plt.ylabel('%s auto-spectrum [%s$^2$s]' % (dof,dim))
    color = plt.cm.rainbow(np.linspace(0,1,len(kwargs)))
    for i, (k, v) in enumerate(kwargs.items()):
        plt.axvline(v, label=k, linestyle='--', color=color[i])
    # Add the response spectrum based on the WADAM RAOs
    if fz_WADAM is not None:
        plt.plot(fz_WADAM, Ss_WADAM, label="WADAM (WF only)")
    plt.title(title)
    ax1 = plt.gca()
    ax2 = ax1.twinx()
    ax2.plot(fw, Sw, label="wave", linestyle='--', color='k')
    ax2.set_ylabel("Wave Spectrum ($m^2s$)", color='k')
    ax2.set_ylim(0)
    l1, h1 = ax1.get_legend_handles_labels()
    l2, h2 = ax2.get_legend_handles_labels()
    plt.legend(l1+l2, h1+h2, loc=0)
    plt.grid()

    if legend is not None:
        plt.legend((legend))
    if filename!=0:
        plt.savefig(filename+'_compautospec')
    if showflag==1:
        plt.show()
    
    plt.close()



#######################################################################################
#######################################################################################

def compareRAOs(wsimo,wmodel,xsimo,xmodel,tsimo,tmodel,filename,rao_diff=None,
    periods_diff=None,diff_label="WADAM",title=None,xmin=0,xmax=30.,
    testno=None,hertz=False,ylabel=""):
    fig, ax1 = plt.subplots()
    rao_sim, fz_sim = rao(xsimo, wsimo, tsimo, smooth=True, bw=0.001)
    rao_test, fz_test = rao(xmodel, wmodel, tmodel, smooth=True, bw=0.001)
    ix = np.nonzero(rao_test)
    ax1.plot(fz_test[ix] if hertz else 1/fz_test[ix], np.abs(rao_test[ix]), label=f"Test {testno if testno is not None else ''}")
    ix = np.nonzero(rao_sim)
    ax1.plot(fz_sim[ix] if hertz else 1/fz_sim[ix], np.abs(rao_sim[ix]), label=f"SIMO")
    if rao_diff is not None: ax1.plot(1/periods_diff if hertz else periods_diff, np.abs(rao_diff), label=diff_label)
    if title is not None: fig.suptitle(title)
    ax1.set_xlabel("Wave " + "frequency [Hz]" if hertz else "period [s]")
    ax1.set_ylabel(ylabel)
    ax1.set_xlim(xmin,xmax)
    ax1.set_ylim(0)
    ax1.legend(loc=0)
    plt.savefig(filename + "_RAO.png")
    plt.close(fig)

def compareRAOsWisting(wsimo,wmodel,xsimo,xmodel,tsimo,tmodel,filename=None,rao_diff=None,
    periods_diff=None,diff_label="WADAM",title=None,xmin=0,xmax=30.,
    testno=None,hertz=False,ylabel="",axs=None, index=0):
    if axs is None:
        fig, ax1 = plt.subplots(ncols=3, figsize=(12,4))
        plt.subplots_adjust(wspace=0.3)
    else:
        fig, ax1 = plt.gcf(), axs
    rao_sim, fz_sim = rao(xsimo, wsimo, tsimo, smooth=True, bw=0.001)
    rao_test, fz_test = rao(xmodel, wmodel, tmodel, smooth=True, bw=0.001)
    ix = np.nonzero(rao_test)
    ax1[index].plot(fz_test[ix] if hertz else 1/fz_test[ix], np.abs(rao_test[ix]), label=f"Test {testno if testno is not None else ''}")
    ix = np.nonzero(rao_sim)
    ax1[index].plot(fz_sim[ix] if hertz else 1/fz_sim[ix], np.abs(rao_sim[ix]), label=f"SIMO")
    if rao_diff is not None: ax1[index].plot(1/periods_diff if hertz else periods_diff, np.abs(rao_diff), label=diff_label)
    if title is not None: fig.suptitle(title)
    ax1[index].set_xlabel("Wave " + "frequency [Hz]" if hertz else "period [s]")
    ax1[index].set_ylabel(ylabel)
    ax1[index].set_xlim(xmin,xmax)
    ax1[index].set_ylim(0)
    if index==0: ax1[index].legend(loc=0)
    if filename is None: return ax1
    plt.savefig(filename + "_RAO.png")
    plt.close(fig)


def compRAO(fz,Ss,fzW,SsW,xmax,label,showflag,filename,axisflag=0,timeflag=0,xmin=0,title=''):

    """specpack.plotRAO
       -----------------

       Plotting of RAO
       
       -----------------

       Input:   fz       - frequency vector [Hz]
                RAO      - RAO to be plotted
                xmax     - max value on x-axis
                showflag - if 1, then figure is shown
                filename - if 0, then nothing is done
                           otherwise, figure is saved as filename_RAO.png
                axisflag - if 0, do not plot RAO for frequencies 
                                 lower than 0.025 Hz
                           if 1, plot RAO for low frequencies as well
                timeflag - if 0, plot against frequency
                           if 1, plot against time
                           default is 0
                xmin     - min frequency in plot, default is 0
                title    - plot title, default is ''
                           

       Output:  Optional show() of figure
                Optional savefig() of figure

       -----------------"""
    
    plt.figure()
    plt.clf()

    if axisflag==0:
        minfreq = 0.025    
        ind_low = plt.find(fz>minfreq)[0]
    elif axisflag==1:
        ind_low = 1

    ind_top = len(Ss)
    if len(plt.find(fz<xmax)): ind_top = plt.find(fz<xmax)[-1]


    if timeflag==0:
        plt.plot(fz[ind_low:ind_top],Ss[ind_low:ind_top],linewidth=2.0,label='Model Test')
        plt.plot(fzW, SsW, label='WAMIT')
        plt.xlabel('Frequency [Hz]')
    elif timeflag==1:
        plt.plot(1/fz[ind_low:ind_top],Ss[ind_low:ind_top],linewidth=2.0,label='Model Test')
        plt.plot(fzW, SsW, label='WAMIT')
        plt.xlabel('Period [s]')

    ymax = max(1.05*Ss[ind_low:ind_top])
    plt.axis([xmin,xmax,0,ymax])
    plt.ylabel(label)
    plt.title(title)
    plt.legend(loc=0)
    plt.grid()    
    if filename!=0:
        plt.savefig(filename+'_RAO')
    if showflag==1:
        plt.show()

    plt.close()




#################


def wave_ts(S, t, seed=None, RAS=False):
    """
    Generates a wave time series from a wave spectrum using random phases, where
    S(w) is a function that returns the spectrums value at w. If RAS is True, random
    amplitudes at each frequency are drawn from the respective Rayleigh distribution

    TODO: Issue to resolve with RAS setup. 
    The difference betweem the correct and applied approach for RAS is one uses a
    factor of sqrt(2/pi) = 0.79 and the other uses a factor sqrt(1/2) = 0.71
    

    Example of use:
    ===============
    from wafo.spectrum.models import Torsethaugen
    from teklib.signal_analysis.specpack import wave_ts

    S = Torsethaugen(Hm0=7, Tp=10) 
    dt = 0.2 # s
    t = np.arange(0,3*60*60*5)*dt
    eta = wave_ts(S, t)
    """    

    # Establish one-sided frequency array
    dt, N, df = t[1] - t[0], len(t), 1/len(t)/(t[1]-t[0])
    fz = np.fft.fftfreq(len(t), d=dt)[:N//2]
    assert(fz[1]-fz[0] == df), "df mismatch"

    # Generate spectral amplitudes 
    Sf = S(2*np.pi*fz)*2*np.pi
    np.random.seed(seed) # Seed option for reproducibility
    phases = np.random.rand(len(Sf))*2*np.pi
    mean_ap = np.sqrt(2*Sf*df)
    mode_ap = np.sqrt(2/np.pi)*mean_ap
    #scale = np.sqrt(4*Sf*df/(4-np.pi)) # Using paper with 2*Sf*df as variance and wiki's scale relation
    ap = np.random.rayleigh(mode_ap) if RAS else mean_ap # this RAS should be correct
    ap = np.random.rayleigh(1/2**0.5*mean_ap) if RAS else mean_ap # but this hits Hs
    #ap = np.random.rayleigh(scale) if RAS else mean_ap # but this hits Hs
    ap = ap*np.exp(1j*phases)

    # ifft back to generate a wave elevation
    eta = np.fft.ifft(ap*len(t), n=len(t)) 



    return eta.real




########################################################################
# program: rainflow.py
# author: Tom Irvine
# Email: tom@vibrationdata.com
# version: 1.4
# date: October 4, 2013
# description:  
#
# Modified version: tike@statoil.com
# Modifications: only input
#
#    - to avoid function calls to tompy
#    - to facilitate command line processing
#    - to provide a calling function for integrating into other programs
#
#  ASTM E 1049-85 (2005) Rainflow Counting Method
#              
########################################################################

import sys
import numpy as np
import pylab as plt
from time import time

########################################################################

class Rainflow:

    def __init__(self):
    
        self.b=[]
        self.ymax=0.
        self.num=0



    @classmethod 
    def rainflow_input(cls,self,data,submean=False):
        self.b=data
        self.num = len(self.b)
        if submean: self.b-=np.mean(self.b)  

   
    @classmethod    
    def rainflow_bins(cls,self,B,kv,hold,sum,nbins):        
        
        L=np.r_[0,np.linspace(0,100,nbins+1)]
        C=np.zeros(nbins+2,'f')

        AverageMean = np.zeros(nbins+2,'f')
        MaxPeak     = np.zeros(nbins+2,'f')
        MinValley   = np.zeros(nbins+2,'f')
        MaxAmp      = np.zeros(nbins+2,'f')
        AverageAmp  = np.zeros(nbins+2,'f')

        for ijk in range (1,nbins+2):
    
            L[ijk]*=self.ymax/100.
    
            MaxPeak[ijk]=-1.0e+20
            MinValley[ijk]= 1.0e+20

            MaxAmp[ijk]=-1.0e+20

        kv-=1


        for i in range (0,kv+1):
             Y=B[i][0]
          
             for ijk in range (nbins,0,-1):
         
                if(Y>=L[ijk] and Y<=L[ijk+1]):
                    
                    C[ijk]+=B[i][1]
            
                    AverageMean[ijk]+=B[i][1]*(B[i][3]+B[i][2])*0.5  # weighted average    

                    if(B[i][3]>MaxPeak[ijk]):   MaxPeak[ijk]   = B[i][3]
                    if(B[i][2]>MaxPeak[ijk]):   MaxPeak[ijk]   = B[i][2]
                    if(B[i][3]<MinValley[ijk]): MinValley[ijk] = B[i][3]
                    if(B[i][2]<MinValley[ijk]): MinValley[ijk] = B[i][2]
                    if(Y>MaxAmp[ijk]): MaxAmp[ijk]=Y

                    AverageAmp[ijk]+=B[i][1]*Y*0.5

                    break

        for ijk in range (1,nbins+2):

            if(C[ijk]>0):
                AverageMean[ijk]/=C[ijk]
                AverageAmp[ijk]/=C[ijk]
              
            if(C[ijk]<0.5):
                AverageMean[ijk]=0.    
                AverageAmp[ijk]=0.       
                MaxPeak[ijk]=0.            
                MinValley[ijk]=0.         
                MaxAmp[ijk]=0.            
        
            MaxAmp[ijk]/=2.


        N = nbins
        midrange  = np.r_[[(L[i]+L[i+1])/2. for i in range(1,nbins+1)]]
        widths    = np.r_[[L[i+1]-L[i] for i in range(1,nbins+1)]]
        numcycles = np.r_[C[1:N+1]]


        print('\n  Amplitude = (peak-valley)/2 \n\n')
#*****************************************************************************

        print("          Range            Cycle       Ave      Max     Ave     Min       Max")
        print("         (units)           Counts      Amp      Amp     Mean    Valley    Peak")

        for i in range (13,0,-1):
            print(("  %8.2f to %8.2f\t%8.1f\t%6.4g\t%6.4g\t%6.4g\t%6.4g\t %6.4g" \
             %(L[i],L[i+1],C[i],AverageAmp[i],MaxAmp[i],AverageMean[i],MinValley[i],MaxPeak[i])))

        yf=self.ymax

        print(("\n\n  Total Cycles = %g  hold=%d  NP=%d ymax=%g" %(sum,hold,self.num,yf)))	

        return midrange, widths, numcycles
            
########################################################################

    def rainflow_core(self, data, nbins, ymax=False):
        
        t0 = time()

        Rainflow.rainflow_input(self, data, False)
                 
########################################################################

        y=np.zeros(self.num,'f')
        a=np.zeros(self.num,'f')
        B=np.zeros((self.num,4),'f')

        y=self.b
        k=0
        a[0]=y[0]

        k=1

        for i in range (1,(self.num-1)):
            
            slope1=(  y[i]-y[i-1])
            slope2=(y[i+1]-y[i])

            if((slope1*slope2)<=0):
                a[k]=y[i]
                k+=1
                 
            
        last_a=k        
        hold=last_a        
        a[k]=y[self.num-1]

        mina=min(a)
        maxa=max(a)


        if sys.version_info[0] == 2:
            nmm=int(maxa-mina)+1

        if sys.version_info[0] == 3:
            nmm=int(maxa-mina)+1

	
        i=0
        j=1

        sum=0
        kv=0

        self.ymax=-1.0e+20

        print (" ")

        aa=a.tolist()

        nkv=0

        print ("\n percent completed \n")

        ijk=0
        LLL=int(0.2*float(hold/2))

        while(1):
        
            Y=abs(aa[i]-aa[i+1])
            X=abs(aa[j]-aa[j+1])
    
            if(X>=Y and Y>0):		
                if(Y>self.ymax):
                    self.ymax=Y
             
                if(i==0):		
                    
                    sum+=0.5
                    B[kv][3]=aa[i+1]
                    B[kv][2]=aa[i]
                    B[kv][1]=0.5
                    B[kv][0]=Y
                    kv+=1
                    aa.pop(i)
                    last_a-=1
                    i=0
                    j=1    
                
                else:      
                    sum+=1
                    B[kv][3]=aa[i+1]
                    B[kv][2]=aa[i]
                    B[kv][1]=1.
                    B[kv][0]=Y
                    kv+=1	

                    aa.pop(i+1)
                    aa.pop(i)                
                
                    i=0
                    j=1

                    last_a-=2
            
                    nkv+=1

                    ijk+=1  
  
                    if(ijk==LLL):
                        pr=(sum/(hold/2))*100.
                        print((" %3.0f" %(pr)))
                        ijk=0               
             	    
            else:
                i+=1
                j+=1
                
            if((j+1)>last_a):
                break                
 
#### 
 
        for i in range (0,last_a):
	
            Y=(abs(aa[i]-aa[i+1]))

            if(Y>0):
		
                sum+=0.5
                B[kv][3]=aa[i+1]
                B[kv][2]=aa[i]
                B[kv][1]=0.5
                B[kv][0]=Y
        
                kv+=1
		
                if(Y>self.ymax):
                   self.ymax=Y
 
        if ymax: self.ymax = ymax
        print((" ymax=%8.4g " %self.ymax))


        midrange, widths, numcycles = Rainflow.rainflow_bins(self,B,kv,hold,sum,nbins)
        return midrange, widths, numcycles

  
###############################################################################
  
class CDI:

    def __init__(self):
    
        pass
   
    def fatigue(self,B,b):   
        
        D=0
        
        kv=B.shape[0]

        for i in range (0,kv):
       
            Y=B[i][0]
            D+=B[i][1]*((Y/2.)**b)
       
        print(("\n D=%8.4g \n" %D))

###############################################################################        




#!/usr/bin/env python
#-------------------------------------------------------------------
# Filename: filter.py
#  Purpose: Various Seismogram Filtering Functions
#   Author: Tobias Megies, Moritz Beyreuther, Yannik Behr
#    Email: tobias.megies@geophysik.uni-muenchen.de
#
# Copyright (C) 2009 Tobias Megies, Moritz Beyreuther, Yannik Behr
#---------------------------------------------------------------------
"""
Various Seismogram Filtering Functions
 
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
 
from numpy import array, where, fft
from scipy.fftpack import hilbert
from scipy.signal import iirfilter, lfilter, remez, convolve, get_window
 
 
def bandpass(data, freqmin, freqmax, df=200, corners=4, zerophase=False):
    """
    Butterworth-Bandpass Filter.
 
    Filter data from freqmin to freqmax using
    corners corners.
 
    :param data: Data to filter, type numpy.ndarray.
    :param freqmin: Pass band low corner frequency.
    :param freqmax: Pass band high corner frequency.
    :param df: Sampling rate in Hz; Default 200.
    :param corners: Filter corners. Note: This is twice the value of PITSA's 
        filter sections
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    [b, a] = iirfilter(corners, [freqmin / fe, freqmax / fe], btype='band',
                       ftype='butter', output='ba')
    if zerophase:
        firstpass = lfilter(b, a, data)
        return lfilter(b, a, firstpass[::-1])[::-1]
    else:
        return lfilter(b, a, data)
 
 
def bandpassZPHSH(data, freqmin, freqmax, df=200, corners=2):
    """
    DEPRECATED. Use :func:`~obspy.signal.filter.bandpass` instead.
    """
    return bandpass(data, freqmin, freqmax, df, corners, zerophase=True)
 
 
def bandstop(data, freqmin, freqmax, df=200, corners=4, zerophase=False):
    """
    Butterworth-Bandstop Filter.
 
    Filter data removing data between frequencies freqmin and freqmax using
    corners corners.
 
    :param data: Data to filter, type numpy.ndarray.
    :param freqmin: Stop band low corner frequency.
    :param freqmax: Stop band high corner frequency.
    :param df: Sampling rate in Hz; Default 200.
    :param corners: Filter corners. Note: This is twice the value of PITSA's 
        filter sections
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    [b, a] = iirfilter(corners, [freqmin / fe, freqmax / fe],
                       btype='bandstop', ftype='butter', output='ba')
    if zerophase:
        firstpass = lfilter(b, a, data)
        return lfilter(b, a, firstpass[::-1])[::-1]
    else:
        return lfilter(b, a, data)
 
 
def bandstopZPHSH(data, freqmin, freqmax, df=200, corners=2):
    """
    DEPRECATED. Use :func:`~obspy.signal.filter.bandstop` instead.
    """
    return bandstop(data, freqmin, freqmax, df, corners, zerophase=True)
 
 
def lowpass(data, freq, df=200, corners=4, zerophase=False):
    """
    Butterworth-Lowpass Filter.
 
    Filter data removing data over certain frequency freq using corners 
    corners.
 
    :param data: Data to filter, type numpy.ndarray.
    :param freq: Filter corner frequency.
    :param df: Sampling rate in Hz; Default 200.
    :param corners: Filter corners. Note: This is twice the value of PITSA's 
        filter sections
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    [b, a] = iirfilter(corners, freq / fe, btype='lowpass', ftype='butter',
                       output='ba')
    if zerophase:
        firstpass = lfilter(b, a, data)
        return lfilter(b, a, firstpass[::-1])[::-1]
    else:
        return lfilter(b, a, data)
 
 
def lowpassZPHSH(data, freq, df=200, corners=2):
    """
    DEPRECATED. Use :func:`~obspy.signal.filter.lowpass` instead.
    """
    return lowpass(data, freq, df, corners, zerophase=True)
 
 
def highpass(data, freq, df=200, corners=4, zerophase=False):
    """
    Butterworth-Highpass Filter.
 
    Filter data removing data below certain frequency freq using corners.
 
    :param data: Data to filter, type numpy.ndarray.
    :param freq: Filter corner frequency.
    :param df: Sampling rate in Hz; Default 200.
    :param corners: Filter corners. Note: This is twice the value of PITSA's 
        filter sections
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    [b, a] = iirfilter(corners, freq / fe, btype='highpass', ftype='butter',
                       output='ba')
    if zerophase:
        firstpass = lfilter(b, a, data)
        return lfilter(b, a, firstpass[::-1])[::-1]
    else:
        return lfilter(b, a, data)
 
 
def highpassZPHSH(data, freq, df=200, corners=2):
    """
    DEPRECATED. Use :func:`~obspy.signal.filter.highpass` instead.
    """
    return highpass(data, freq, df, corners, zerophase=True)
 
 
def envelope(data):
    """
    Envelope of a function.
 
    Computes the envelope of the given function. The envelope is determined by
    adding the squared amplitudes of the function and it's Hilbert-Transform 
    and then taking the squareroot. 
    (See Kanasewich: Time Sequence Analysis in Geophysics)
    The envelope at the start/end should not be taken too seriously.
 
    :param data: Data to make envelope of, type numpy.ndarray.
    :return: Envelope of input data.
    """
    hilb = hilbert(data)
    data = pow(pow(data, 2) + pow(hilb, 2), 0.5)
    return data
 
 
def remezFIR(data, freqmin, freqmax, samp_rate=200):
    """
    The minimax optimal bandpass using Remez algorithm. Zerophase bandpass?
 
    Finite impulse response (FIR) filter whose transfer function minimizes
    the maximum error between the desired gain and the realized gain in the
    specified bands using the remez exchange algorithm
    """
    # Remez filter description
    # ========================
    #
    # So, let's go over the inputs that you'll have to worry about.
    # First is numtaps. This parameter will basically determine how good your
    # filter is and how much processor power it takes up. If you go for some
    # obscene number of taps (in the thousands) there's other things to worry
    # about, but with sane numbers (probably below 30-50 in your case) that is
    # pretty much what it affects (more taps is better, but more expensive
    #         processing wise). There are other ways to do filters as well
    # which require less CPU power if you really need it, but I doubt that you
    # will. Filtering signals basically breaks down to convolution, and apple
    # has DSP libraries to do lightning fast convolution I'm sure, so don't
    # worry about this too much. Numtaps is basically equivalent to the number
    # of terms in the convolution, so a power of 2 is a good idea, 32 is
    # probably fine.
    #
    # bands has literally your list of bands, so you'll break it up into your
    # low band, your pass band, and your high band. Something like [0, 99, 100,
    # 999, 1000, 22049] should work, if you want to pass frequencies between
    # 100-999 Hz (assuming you are sampling at 44.1 kHz).
    #
    # desired will just be [0, 1, 0] as you want to drop the high and low
    # bands, and keep the middle one without modifying the amplitude.
    #
    # Also, specify Hz = 44100 (or whatever).
    #
    # That should be all you need; run the function and it will spit out a list
    # of coefficients [c0, ... c(N-1)] where N is your tap count. When you run
    # this filter, your output signal y[t] will be computed from the input x[t]
    # like this (t-N means N samples before the current one):
    #
    # y[t] = c0*x[t] + c1*x[t-1] + ... + c(N-1)*x[t-(N-1)]
    #
    # After playing around with remez for a bit, it looks like numtaps should 
    # be above 100 for a solid filter. See what works out for you. Eventually, 
    # take those coefficients and then move them over and do the convolution 
    # in C or whatever. Also, note the gaps between the bands in the call to 
    # remez. You have to leave some space for the transition in frequency 
    # response to occur, otherwise the call to remez will complain.
    #
    # SRC: # http://episteme.arstechnica.com/eve/forums/a/tpc/f/6330927813/m/175006289731
    # See also:
    # http://aspn.activestate.com/ASPN/Mail/Message/scipy-dev/1592174
    # http://aspn.activestate.com/ASPN/Mail/Message/scipy-dev/1592172
 
    #take 10% of freqmin and freqmax as """corners"""
    flt = freqmin - 0.1 * freqmin
    fut = freqmax + 0.1 * freqmax
    #bandpass between freqmin and freqmax
    filt = remez(50, array([0, flt, freqmin, freqmax, fut, samp_rate / 2 - 1]),
                 array([0, 1, 0]), Hz=samp_rate)
    return convolve(filt, data)
 
 
def lowpassFIR(data, freq, samp_rate=200, winlen=2048):
    """
    FIR-Lowpass Filter
 
    Filter data by passing data only below a certain frequency.
 
    :param data: Data to filter, type numpy.ndarray.
    :param freq: Data below this frequency pass.
    :param samprate: Sampling rate in Hz; Default 200.
    :param winlen: Window length for filter in samples, must be power of 2; 
        Default 2048
    :return: Filtered data.
    """
    # There is not currently an FIR-filter design program in SciPy.  One 
    # should be constructed as it is not hard to implement (of course making 
    # it generic with all the options you might want would take some time).
    # 
    # What kind of window are you currently using?
    # 
    # For your purposes this is what I would do:
    # SRC: Travis Oliphant
    # http://aspn.activestate.com/ASPN/Mail/Message/scipy-user/2009409]
    #
    #winlen = 2**11 #2**10 = 1024; 2**11 = 2048; 2**12 = 4096
    #give frequency bins in Hz and sample spacing
    w = fft.fftfreq(winlen, 1 / float(samp_rate))
    #cutoff is low-pass filter
    myfilter = where((abs(w) < freq), 1., 0.)
    #ideal filter
    h = fft.ifft(myfilter)
    beta = 11.7
    #beta implies Kaiser
    myh = fft.fftshift(h) * get_window(beta, winlen)
    return convolve(abs(myh), data)[winlen / 2:-winlen / 2]
 
